Require Import FJ_tactics.
Require Import List.
Require Import Functors.
Require Import Names.
Require Import FunctionalExtensionality.
Require Import MonadLib.

Section Exception.

  (* ============================================== *)
  (* EXPRESSIONS                                    *)
  (* ============================================== *)

  (* Functor for Types. *)
  Variable D : Set -> Set.
  Context {Fun_D : Functor D}.

  Inductive ExceptE (A : Set) : Set :=
  | Throw : DType D -> ExceptE A
  | Catch : A -> (unit -> A) -> ExceptE A.

  Definition ExceptE_fmap (B C : Set) (f : B -> C) (e : ExceptE B) : ExceptE C :=
    match e with
      | Throw n => Throw _ n
      | Catch e' h => Catch _ (f e') (fun n => f (h n))
    end.

  Global Instance ExceptE_Functor : Functor ExceptE :=
    {| fmap := ExceptE_fmap |}.
    destruct a; reflexivity.
    (* fmap id *)
    destruct a; simpl; auto.
    unfold id; apply f_equal; apply functional_extensionality; auto.
  Defined.

  Variable F : Set -> Set.
  Context {Fun_F : Functor F}.
  Definition Exp := Exp F.
  Context {Sub_ExceptE_F : ExceptE :<: F}.

  (* Constructor + Universal Property. *)
   Context {WF_Sub_ExceptE_F : WF_Functor _ _ Sub_ExceptE_F _ _}.
  Definition throw'' (t : DType D) : Exp := inject' (Throw _ t).
  Definition throw'  (t : DType D) : Fix F := proj1_sig (throw'' t).
  Global Instance UP'_lit (t : DType D) :
    Universal_Property'_fold (throw' t) := proj2_sig (throw'' t).

  Definition catch'' (e : Exp) (h : unit -> Exp) : Exp :=  inject' (Catch _ e h).

  Definition catch' (e : Fix F) (h : unit -> Fix F)
    {UP'_e : Universal_Property'_fold e}
    {UP'_h : forall n, Universal_Property'_fold (h n)}
    : Fix F := proj1_sig (catch'' (exist _ _ UP'_e) (fun n => exist _ _ (UP'_h n))).

   Global Instance UP'_add {e : Fix F} {h : unit -> Fix F}
    {UP'_e : Universal_Property'_fold e}
    {UP'_h : forall n, Universal_Property'_fold (h n)}
    : Universal_Property'_fold (catch' e h) :=
    proj2_sig (catch'' (exist _ _ UP'_e) (fun n => exist _ _ (UP'_h n))).

  (* Induction Principles for Arith. *)
  Definition ind_alg_ExceptE
    (P : forall e : Fix F, Universal_Property'_fold e -> Prop)
    (H : forall n, UP'_P P (throw' n))
    (H0 : forall (e : Fix F) (h : unit -> Fix F)
      (IHe : UP'_P P e)
      (IHh : forall n, UP'_P P (h n)),
      UP'_P P (@catch' _ _ (proj1_sig IHe) (fun n => proj1_sig (IHh n))))
      (e : ExceptE (sig (UP'_P P)))
        :
        sig (UP'_P P) :=
    match e with
      | Throw n => exist _ (throw' n) (H n)
      | Catch e h => exist (UP'_P P) _
        (H0 _ _ (proj2_sig e) (fun n => proj2_sig (h n)))
    end.

  Lemma ind_PAlg_ExceptE (Name : Set)
    (P : forall e : Fix F, Universal_Property'_fold e -> Prop)
    (H : forall n, UP'_P P (throw' n))
    (H0 : forall e h
      (IHe : UP'_P P e)
      (IHh : forall n, UP'_P P (h n)),
      UP'_P P (@catch' e h (proj1_sig IHe) (fun n => proj1_sig (IHh n))))
    :
    PAlgebra Name ExceptE F (UP'_P P).
    econstructor 1 with (p_algebra := ind_alg_ExceptE P H H0).
    destruct e; simpl.
    unfold throw', throw''; simpl; rewrite wf_functor; reflexivity.
    unfold catch', catch''; simpl; rewrite wf_functor; reflexivity.
  Qed.

  Definition indD_alg_ExceptE
    (P : forall e : Fix F, Universal_Property'_fold e -> Set)
    (H : forall n, UP'_SP P (throw' n))
    (H0 : forall (e : Fix F) (h : unit -> Fix F)
      (IHe : UP'_SP P e)
      (IHh : forall n, UP'_SP P (h n)),
      UP'_SP P (@catch' _ _ (projT1 IHe) (fun n => projT1 (IHh n))))
      (e : ExceptE (sigT (UP'_SP P)))
        :
        sigT (UP'_SP P) :=
    match e with
      | Throw n => existT _ (throw' n) (H n)
      | Catch e h => existT (UP'_SP P) _
        (H0 _ _ (projT2 e) (fun n => projT2 (h n)))
    end.

  Lemma ind_DAlg_ExceptE (Name : Set)
    (P : forall e : Fix F, Universal_Property'_fold e -> Set)
    (H : forall n, UP'_SP P (throw' n))
    (H0 : forall (e : Fix F) (h : unit -> Fix F)
      (IHe : UP'_SP P e)
      (IHh : forall n, UP'_SP P (h n)),
      UP'_SP P (@catch' _ _ (projT1 IHe) (fun n => projT1 (IHh n))))
    :
    DAlgebra Name ExceptE F (UP'_SP P).
    econstructor 1 with (d_algebra := indD_alg_ExceptE P H H0).
    destruct e; simpl.
    unfold throw', throw''; simpl; rewrite wf_functor; reflexivity.
    unfold catch', catch''; simpl; rewrite wf_functor; reflexivity.
  Qed.

  Definition ind2_alg_ExceptE
    {E E' : Set -> Set}
    {Fun_E : Functor E}
    {Fun_E' : Functor E'}
    {Sub_Arith_E : ExceptE :<: E}
    {Sub_Arith_E' : ExceptE :<: E'}
    (P : forall e : (Fix E) * (Fix E'),
      Universal_Property'_fold (fst e) /\ Universal_Property'_fold (snd e) -> Prop)
    (H : forall n, UP'_P2 P (inject (Throw _ n), inject (Throw _ n)))
    (H0 : forall e h
      (IHe : UP'_P2 P e)
      (IHh : forall n, UP'_P2 P (h n)),
      UP'_P2 P (inject (Catch _ (exist _ _ (proj1 (proj1_sig IHe)))
        (fun n => exist _ _ (proj1 (proj1_sig (IHh n))))),
      inject (Catch _ (exist _ _ (proj2 (proj1_sig IHe)))
        (fun n => exist _ _ (proj2 (proj1_sig (IHh n)))))))
    (e : ExceptE (sig (UP'_P2 P)))
        :
        sig (UP'_P2 P) :=
    match e with
      | Throw n => exist _ _ (H n)
      | Catch e h => exist (UP'_P2 P) _
        (H0 _ _ (proj2_sig e) (fun n => proj2_sig (h n)))
    end.

  Lemma ind_P2Alg_ExceptE (Name : Set)
    (E E' : Set -> Set)
    {Fun_E : Functor E}
    {Fun_E' : Functor E'}
    {Sub_Arith_E : ExceptE :<: E}
    {Sub_Arith_E' : ExceptE :<: E'}
    {WF_Fun_E : WF_Functor _ _ _ _ Fun_E}
    {WF_Fun_E' : WF_Functor _ _ _ _ Fun_E'}
    (P : forall e : (Fix E) * (Fix E'),
      Universal_Property'_fold (fst e) /\ Universal_Property'_fold (snd e) -> Prop)
    (H : forall n, UP'_P2 P (inject (Throw _ n), inject (Throw _ n)))
    (H0 : forall e h
      (IHe : UP'_P2 P e)
      (IHh : forall n, UP'_P2 P (h n)),
      UP'_P2 P (inject (Catch _ (exist _ _ (proj1 (proj1_sig IHe)))
        (fun n => exist _ _ (proj1 (proj1_sig (IHh n))))),
      inject (Catch _ (exist _ _ (proj2 (proj1_sig IHe)))
        (fun n => exist _ _ (proj2 (proj1_sig (IHh n)))))))
    :
    P2Algebra Name ExceptE E E' (UP'_P2 P).
    econstructor 1 with (p2_algebra := ind2_alg_ExceptE P H H0);
    destruct e; simpl; unfold inject; simpl; rewrite wf_functor; reflexivity.
  Qed.

  (* ============================================== *)
  (* VALUES                                         *)
  (* ============================================== *)

  (* No Values for exceptions *)

   Variable V : Set -> Set.
   Context {Fun_V : Functor V}.
   Definition Value := Value V.

   Context {Sub_StuckValue_V : StuckValue :<: V}.
   Definition stuck' : nat -> Value := stuck' _.
   Definition stuck : nat -> Fix V := stuck _.

  (* Typing Reference Expressions. *)

   Variable (MT : Set -> Set).
   Context {Fun_MT : Functor MT}.
   Context {Mon_MT : Monad MT}.
   Context {Fail_MT : FailMonad MT}.

  Context {eq_DType_DT : forall T, FAlgebra eq_DTypeName T (eq_DTypeR D) D}.

  Definition ExceptE_typeof (R : Set) (rec : R -> typeofR D MT)
     (e : ExceptE R) : typeofR D MT :=
     match e with
       | Throw t => return_ t
       | Catch e' h => do te <- (rec e');
                       do th <- (rec (h tt));
                         if (eq_DType D (proj1_sig te) th) then
                           return_ te else fail
     end.

  Global Instance MAlgebra_typeof_ExceptE T:
    FAlgebra TypeofName T (typeofR D MT) ExceptE :=
    {| f_algebra := ExceptE_typeof T|}.

  (* ============================================== *)
  (* EVALUATION                                     *)
  (* ============================================== *)

  Context {Sub_BotValue_V : BotValue :<: V}.

   (* Evaluation Algebra for Arithemetic Expressions. *)

  Variable (ME : Set -> Set). (* Evaluation Monad. *)
  Context {Fun_M : Functor ME}.
  Context {Mon_M : Monad ME}.
  Context {Exception_ME : Exception ME unit}.
  Context {Fail_ME : FailMonad ME}.
  (* Monadic Evaluation Algebra for Exception Expressions. *)

  Definition ExceptE_evalM (R : Set) (rec : R -> evalMR V ME)
     (e : ExceptE R) : evalMR V ME :=
       match e with
         | Throw t => throw (M := ME) tt
         | Catch e h => catch (rec e) (fun n => rec (h n))
       end.

  Global Instance MAlgebra_evalM_ExceptE T :
    FAlgebra EvalName T (evalMR V ME) ExceptE :=
    {| f_algebra := ExceptE_evalM T|}.

  (* ============================================== *)
  (* PRETTY PRINTING                                *)
  (* ============================================== *)

  (* Pretty Printing Functions*)

  Require Import Ascii.
  Require Import String.

  Global Instance MAlgebra_ExpPrint_ExceptE T :
    FAlgebra ExpPrintName T (ExpPrintR) ExceptE :=
    {| f_algebra := fun rec e => match e with
      | Throw n => fun i => append "(throw)" EmptyString (*(String (ascii_of_nat (n + 48)) EmptyString) ++ ")" *)
      | Catch m n => fun i => append "(catch" ((rec m i) ++ " with " ++ (rec (n tt) (i + 1)) ++ ")")
    end |}.

  (* ============================================== *)
  (* CONTINUITY Of MONADIC EVALUATION               *)
  (* ============================================== *)

  Context {evalM_F : FAlgebra EvalName Exp (evalMR V ME) F}.
  Context {WF_eval_F : @WF_FAlgebra EvalName _ _ ExceptE F
    Sub_ExceptE_F (MAlgebra_evalM_ExceptE _) evalM_F}.

   (* Continuity of Monadic Evaluation. *)

  Context {WF_SubBotValue_V : WF_Functor BotValue V Sub_BotValue_V Bot_Functor Fun_V}.
  Context {SV : (SubValue_i V -> Prop) -> SubValue_i V -> Prop}.
  Context {Sub_SV_refl_SV : Sub_iFunctor (SubValue_refl V) SV}.
  Variable SVM : (SubValueM_i V ME -> Prop) -> SubValueM_i V ME -> Prop.
  Context {iFun_SVM : iFunctor SVM}.
  Context {Sub_SVM_Base_SVM : Sub_iFunctor (SubValueM_base V ME) SVM}.

    (** Subvalue for the Exception Monad. **)

    Inductive SubValueM_Catch (A : SubValueM_i V ME -> Prop) : SubValueM_i V ME -> Prop :=
    | SVM_Throw : forall t,
      SubValueM_Catch A (mk_SubValueM_i _ _ (throw t) (throw t))
    | SVM_Catch : forall me me' h h' (k k' : Names.Value V -> _),
      A (mk_SubValueM_i _ _ (me >>= k) (me' >>= k')) ->
      (forall (i : unit), A (mk_SubValueM_i _ _ (h i >>= k) (h' i >>= k'))) ->
      SubValueM_Catch A (mk_SubValueM_i _ _ (catch me h >>= k) (catch me' h' >>= k')).

    Definition ind_alg_SVM_Catch (P : SubValueM_i V ME -> Prop)
      (H_Throw : forall t, P (mk_SubValueM_i _ _ (throw t) (throw t)))
      (H_Catch : forall me me' h h' k k',
        P (mk_SubValueM_i _ _ (me >>= k) (me' >>= k')) ->
        (forall (i : unit), P (mk_SubValueM_i _ _ (h i >>= k) (h' i >>= k'))) ->
        P (mk_SubValueM_i _ _ (catch me h >>= k) (catch me' h' >>= k')))
      (i : SubValueM_i V ME)
      (e : SubValueM_Catch P i)
      :
      P i :=
      match e in SubValueM_Catch _ i return P i with
        | SVM_Throw t => H_Throw t
        | SVM_Catch me me' h h' k k' P_me_me' P_k_k' =>
          H_Catch me me' h h' k k' P_me_me' P_k_k'
      end.

    Definition SVM_Catch_ifmap
      (A B : SubValueM_i V ME -> Prop)
      (i : SubValueM_i V ME)
      (f : forall i, A i -> B i)
      (SVM_a : SubValueM_Catch A i) : SubValueM_Catch B i :=
      match SVM_a in SubValueM_Catch _ i return SubValueM_Catch B i with
        | SVM_Throw t => SVM_Throw _ t
        | SVM_Catch me me' h h' k k' A_me_me' A_h_h' =>
          SVM_Catch B me me' h h' k k' (f _ A_me_me')
            (fun i => f _ (A_h_h' i))
      end.

    Global Instance iFun_SVM_Catch : iFunctor SubValueM_Catch.
      constructor 1 with (ifmap := SVM_Catch_ifmap).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; auto;
      unfold id; apply f_equal; apply functional_extensionality_dep;
        auto.
    Defined.

    Context {Sub_SVM_Catch_SVM : Sub_iFunctor SubValueM_Catch SVM}.

    Global Instance SVM_Bind_Exception :
      iPAlgebra SVM_Bind_Name (SVM_Bind_P V ME SVM) SubValueM_Catch.
    Proof.
      econstructor; unfold iAlgebra; intros; apply ind_alg_SVM_Catch;
        eauto; unfold SVM_Bind_P; simpl; intros.
      (* SVM_Get *)
      repeat rewrite <- associativity; apply (inject_i (subGF := Sub_SVM_Catch_SVM));
        repeat rewrite bind_throw; constructor 1; intros; eapply H0; eauto.
      (* SVM_Return *)
      unfold wbind; repeat rewrite <- associativity; apply (inject_i (subGF := Sub_SVM_Catch_SVM));
        constructor 2; intros; repeat rewrite associativity.
      eapply H0; eauto.
      eapply H1; eauto.
    Qed.

  (* Throw case. *)

  Ltac WF_FAlg_rewrite := repeat rewrite wf_functor; simpl;
    repeat rewrite out_in_fmap; simpl;
      repeat rewrite wf_functor; simpl;
        repeat rewrite wf_algebra; simpl.

    Lemma bevalM_continuous_Exp_H : forall n,
      UP'_SP (bevalM_continuous_Exp_P V F ME SVM) (throw' n).
      unfold bevalM_continuous_Exp_P; intros; econstructor; intros.
      auto with typeclass_instances.
      unfold Names.bevalM, mfold; simpl; rewrite H1; unfold throw'.
      WF_FAlg_rewrite.
      apply (inject_i (subGF := Sub_SVM_Catch_SVM)); econstructor; simpl; eauto.
    Qed.

    (*Lemma bevalM_continuous_Exp_H : forall n,
      UP'_SP (bevalM_continuous_Exp_P V F ME SV) (throw' n).
      unfold bevalM_continuous_Exp_P; intros; econstructor; intros.
      auto with typeclass_instances.
      unfold Names.bevalM, mfold; simpl; rewrite H1; unfold throw'.
      WF_FAlg_rewrite.
      exists (throw n); rewrite bind_throw, fmap_m, bind_throw;
        auto.
    Qed. *)

    (* Catch case. *)

  Lemma bevalM_continuous_Exp_H0 : forall (e : Fix F) (h : unit -> Fix F)
      (IHe : UP'_SP (bevalM_continuous_Exp_P V F ME SVM) e)
      (IHh : forall n, UP'_SP (bevalM_continuous_Exp_P V F ME SVM) (h n)),
      UP'_SP (bevalM_continuous_Exp_P V F ME SVM)
      (@catch' _ _ (projT1 IHe) (fun n => projT1 (IHh n))).
      unfold bevalM_continuous_Exp_P; intros.
      econstructor; intros; eauto with typeclass_instances.
      unfold Names.bevalM, mfold, catch'; simpl; rewrite H1; unfold catch'; simpl.
      WF_FAlg_rewrite.
      rewrite right_unit with (a := catch _ _).
      rewrite right_unit with (a := catch (@boundedFix_UP _ _ _ n
        _ (@fail _ _ _ Fail_ME (@Names.Value V Fun_V)) _) _).
      destruct IHe as [e_UP' IHe].
      repeat erewrite bF_UP_in_out.
      generalize (H (exist _ _ e_UP') (exist _ _ e_UP') n H0 (refl_equal _)); intros SubM_em_en.
      generalize (fun n' => H (exist _ _ (projT1 (IHh n'))) (exist _ _ (projT1 (IHh n')))
        n H0 (refl_equal _)); intros SubM_hm_hn.
      unfold bevalM in *|-*; apply (inject_i (subGF := Sub_SVM_Catch_SVM));
        econstructor; simpl; repeat rewrite <- right_unit; eauto.
      intros; repeat rewrite <- right_unit;
        repeat erewrite bF_UP_in_out; eapply SubM_hm_hn.
    Qed.

    Global Instance Exception_bevalM_continuous_Exp  :
      DAlgebra EC_ExpName ExceptE F (UP'_SP (bevalM_continuous_Exp_P V F ME SVM)).
    Proof.
      apply ind_DAlg_ExceptE.
      apply bevalM_continuous_Exp_H.
      apply bevalM_continuous_Exp_H0.
    Qed.

  (* ============================================== *)
  (* WELL-FORMED MONADIC EXCEPTIONS               *)
  (* ============================================== *)

    (** SuperFunctor for Well-Formed Monadic Value Relation with
       Typing Environment. **)

  Section WFValue_Sec.

    Variable TypContext : Set.
    Context {TypContextCE : ConsExtensionC TypContext}.

    Inductive WFValueM_Except (A : WFValueM_i D V MT ME TypContext -> Prop) : WFValueM_i D V MT ME TypContext -> Prop :=
    | WFVM_Throw : forall Sigma T,
      WFValueM_Except A (mk_WFValueM_i D V MT ME _ Sigma (throw tt) T)
    | WFVM_Catch : forall (B C : Set) Sigma e' h mte mth
      (k : B -> ME Value) (kT : C -> DType D -> MT (DType D))
      (kT' : MT C)
      (kt_eq : forall (T T' : DType D), proj1_sig T = proj1_sig T' ->
        fmap (@proj1_sig _ _) (do U <- kT'; kT U T) = fmap (@proj1_sig _ _) (do U <- kT'; kT U T')),
      A (mk_WFValueM_i D V MT ME _ Sigma (e' >>= k) (do U <- kT'; mte >>= kT U)) ->
      (forall t Sigma', ConsExtension Sigma' Sigma -> 
        A (mk_WFValueM_i D V MT ME _ Sigma' ((h t) >>= k) (do U <- kT'; mth >>= kT U))) ->
      WFValueM_Except A (mk_WFValueM_i D V MT ME _ Sigma ((catch e' h) >>= k)
        (do U <- kT'; (do te <- mte; do th <- mth; if (eq_DType D (proj1_sig te) th) then (return_ te) else fail) >>= kT U)).

    Definition ind_alg_WFVM_Except (P : WFValueM_i D V MT ME _ -> Prop)
      (H_Throw' : forall Sigma T,
        P (mk_WFValueM_i _ _ _ _ _ Sigma (throw tt) T))
      (H_Catch' : forall A C Sigma e' h mte mth k kT kT' kt_eq,
      P (mk_WFValueM_i D V MT ME _ Sigma (e' >>= k) (do U <- kT'; mte >>= kT U)) ->
      (forall t Sigma', ConsExtension Sigma' Sigma -> 
        P (mk_WFValueM_i D V MT ME _ Sigma' ((h t) >>= k) (do U <- kT'; mth >>= kT U))) ->
      P (mk_WFValueM_i D V MT ME _ Sigma ((catch e' h) >>= k) _))
      (i : WFValueM_i D V MT ME _)
      (e : WFValueM_Except P i)
      :
      P i :=
      match e in WFValueM_Except _ i return P i with
        | WFVM_Throw Sigma T => H_Throw' Sigma T
        | WFVM_Catch A C Sigma e' h mte mth k kT kT' kT_eq P_e' P_h =>
          H_Catch' A C Sigma e' h mte mth k kT kT' kT_eq P_e' P_h
      end.

    Definition WFVM_Except_ifmap
      (A B : WFValueM_i D V MT ME _ -> Prop)
      (i : WFValueM_i D V MT ME _)
      (f : forall i, A i -> B i)
      (WFVM_a : WFValueM_Except A i)
      :
      WFValueM_Except B i :=
      match WFVM_a in WFValueM_Except _ i return WFValueM_Except B i with
        | WFVM_Throw Sigma T => WFVM_Throw B Sigma T
        | WFVM_Catch C D Sigma e' h mte mth k kT kT' kT_eq A_e' A_h =>
          WFVM_Catch B C D Sigma e' h mte mth k kT kT' kT_eq (f _ A_e')
          (fun t Sigma' Sig_sub_Sig' => f _ (A_h t Sigma' Sig_sub_Sig'))
      end.

    Global Instance iFun_wfvm_Except : iFunctor WFValueM_Except.
      constructor 1 with (ifmap := WFVM_Except_ifmap).
      (* ifmap_fusion *)
      destruct a; simpl; intros; reflexivity.
      (* ifmap_id *)
      destruct a; simpl; intros; auto;
        apply f_equal; repeat (apply functional_extensionality_dep; intros); auto.
    Defined.

  End WFValue_Sec.

    Variable TypContext : Set.
    Context {TypContextCE : ConsExtensionC TypContext}.
    Variable WFV : (WFValue_i D V TypContext -> Prop) -> WFValue_i D V TypContext -> Prop.
    Variable funWFV : iFunctor WFV.

    Variable WFVM : (WFValueM_i D V MT ME TypContext -> Prop) -> WFValueM_i D V MT ME TypContext -> Prop.
    Definition WFValueM := iFix WFVM.
    Definition WFValueMC Sigma v T:= WFValueM (mk_WFValueM_i D V MT ME _ Sigma v T).
    Variable funWFVM : iFunctor WFVM.

    Context {Sub_WFVM_Base_WFVM : Sub_iFunctor (WFValueM_base D V MT ME TypContext WFV) WFVM}.

    Context {eq_DType_eq_DT : PAlgebra eq_DType_eqName D D (UP'_P (eq_DType_eq_P D))}.
    Context {eq_DType_neq_D : PAlgebra eq_DType_neqName D D (UP'_P (eq_DType_neq_P D))}.

    Context {Typeof_F : forall T, FAlgebra TypeofName T (typeofR D MT) F}.
    Context {WF_Typeof_F : forall T, @WF_FAlgebra TypeofName _ _ ExceptE F
      Sub_ExceptE_F (MAlgebra_typeof_ExceptE _) (Typeof_F T)}.

    Context {evalM_F' : forall T, FAlgebra EvalName T (evalMR V ME) F}.
    Context {WF_eval_F' : forall T, @WF_FAlgebra EvalName _ _ ExceptE F
      Sub_ExceptE_F (MAlgebra_evalM_ExceptE _) (evalM_F' T)}.

    Context {Sub_WFVM_Except_WFVM : Sub_iFunctor (WFValueM_Except TypContext) WFVM}.

  Lemma ExceptE_Soundness_H : forall n,
    UP'_SP (eval_soundness_P D V F MT ME _ WFVM) (throw' n).
  Proof.
    unfold eval_soundness_P; intros; econstructor; intros.
    auto with typeclass_instances.
    unfold evalM, typeof, mfold; simpl; unfold throw'; simpl; unfold in_t.
    WF_FAlg_rewrite.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor.
  Qed.

  Lemma eval_soundness_H0 (typeof_R eval_R : Set)
      typeof_rec eval_rec
      :
      forall (eT : typeof_R) (hT : unit -> typeof_R)
        (eE : eval_R) (hE : unit -> eval_R) Sigma
        (IHe : WFValueMC Sigma
          (eval_rec eE) (typeof_rec eT))
        (IHh : forall n Sigma', ConsExtension Sigma' Sigma -> 
          WFValueMC Sigma' (eval_rec (hE n)) (typeof_rec (hT n))),
        WFValueMC Sigma
        (ExceptE_evalM _ eval_rec (Catch _ eE hE))
        (ExceptE_typeof _ typeof_rec (Catch _ eT hT)).
    Proof.
      simpl; intros.
      rewrite right_unit with (a := catch _ _).
      rewrite (right_unit (M := MT)).
      rewrite (left_unit (M := MT) (A := unit) tt
        (fun _ => (do te <- typeof_rec eT;
                   do th <- typeof_rec (hT tt);
                     (if eq_DType D (proj1_sig te) th then return_ te else fail)) >>= return_)).
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2.
      intros; repeat rewrite <- left_unit, fmap_return; congruence.
      rewrite <- left_unit, <- right_unit, <- right_unit.
      apply IHe.
      intros; rewrite <- right_unit, <- left_unit, <- right_unit;
        destruct t; eapply IHh; eauto.
    Qed.

  Lemma ExceptE_Soundness_H0 : forall (e : Fix F) (h : unit -> Fix F)
    (IHe : UP'_SP (eval_soundness_P D V F MT ME _ WFVM) e)
    (IHh : forall n, UP'_SP (eval_soundness_P D V F MT ME _ WFVM) (h n)),
    UP'_SP (eval_soundness_P D V F MT ME _ WFVM)
    (@catch' _ _ (projT1 IHe) (fun n => projT1 (IHh n))).
  Proof.
    unfold eval_soundness_P; intros; econstructor; intros.
    auto with typeclass_instances.
    unfold evalM, typeof, mfold; simpl; unfold catch'; simpl; unfold in_t.
    WF_FAlg_rewrite.
    destruct IHe as [e_UP' IHe].
    rewrite right_unit with (a := catch _ _).
    rewrite (right_unit (M := MT)).
    rewrite (left_unit (M := MT) (A := unit) tt
      (fun _ => (do te <- mfold (typeofR D MT) (fun R : Set => f_algebra (FAlgebra := Typeof_F R)) e;
        do th <- mfold (typeofR D MT) (fun R : Set => f_algebra (FAlgebra := Typeof_F R)) (h tt);
          (if eq_DType D (proj1_sig te) th then return_ te else fail)) >>=
      return_)).
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2.
    intros; repeat rewrite <- left_unit, fmap_return; congruence.
    rewrite <- left_unit, <- right_unit, <- right_unit.
    apply IHe.
    intros; rewrite <- right_unit, <- left_unit, <- right_unit; destruct (IHh t) as [ht_UP' IHht].
    destruct t; eapply IHht.
  Qed.

  Global Instance ExceptE_eval_Soundness :
    DAlgebra ES_ExpName ExceptE F (UP'_SP (eval_soundness_P D V F MT ME _ WFVM)).
  Proof.
    apply ind_DAlg_ExceptE.
    apply ExceptE_Soundness_H.
    apply ExceptE_Soundness_H0.
  Qed.

    Lemma catch_msimp : forall (M : Set -> Set) `{Monad M} {A B C D : Set}
      (mta : M A) (mtb : M B) (mtc : A -> B -> M C) (mtd : A -> C -> M D),
      do a <- mta;
        do c <- (mtb >>= mtc a);
      mtd a c =
      do a <- mta;
        do b <- mtb;
          do c <- mtc a b;
            mtd a c.
    Proof.
      intros; apply f_equal; apply functional_extensionality; intros;
        repeat rewrite <- associativity; auto.
    Qed.

    Lemma kt_eq_msimp : forall (M : Set -> Set) `{Monad M} {A B C D : Set}
      (mta : M A) (mtb : A -> M B) (f : C -> M D) (g : B -> C),
      do a <- mta;
        do c <- (mtb a);
          f (g c) =
          fmap g (mta >>= mtb) >>= f.
    Proof.
      intros; rewrite associativity.
      rewrite eq_bind_fmap with (g := g) (f' := f);
        [rewrite Eta with (f := f); auto | auto].
    Qed.

    Lemma curry_msimp : forall (M : Set -> Set) `{Monad M} {A B C : Set}
      (mta : M A) (mtb : A -> M B) (f : A -> B -> M C),
      do a <- mta;
        do b <- mtb a;
          f a b =
          do ab <- (do a <- mta; do b <- mtb a; return_ (a, b)); f (fst ab) (snd ab).
    Proof.
      intros; rewrite <- associativity; apply f_equal; apply functional_extensionality;
        intros; rewrite <- associativity; apply f_equal; apply functional_extensionality;
          intros; rewrite <- left_unit; auto.
    Qed.

    Lemma curry3_msimp : forall (M : Set -> Set) `{Monad M} {A B C D : Set}
      (mta : M A) (mtb : A -> M B) (mtc : A -> B -> M C) (f : A -> B -> C -> M D),
      do a <- mta;
        do b <- mtb a;
          do c <- mtc a b;
            f a b c =
            do abc <- (do a <- mta; do b <- mtb a; do c <- mtc a b; return_ (a, (b, c)));
              f (fst abc) (fst (snd abc)) (snd (snd abc)).
    Proof.
      intros; rewrite <- associativity; apply f_equal; apply functional_extensionality;
        intros; rewrite <- associativity; apply f_equal; apply functional_extensionality;
          intros; rewrite <- associativity; apply f_equal; apply functional_extensionality;
            intros; rewrite <- left_unit; auto.
    Qed.

    Lemma curry3_msimp' : forall (M : Set -> Set) `{Monad M} {A B C D : Set}
      (mta : M A) (mtb : A -> M B) (mtc : prod A B -> M C) (f : A -> B -> C -> M D),
      do ab <- (do a <- mta; do b <- mtb a; return_ (a, b)); do c <- mtc ab;
        f (fst ab) (snd ab) c =
        do abc <- (do a <- mta; do b <- mtb a; do c <- mtc (a, b); return_ (a, (b, c)));
          f (fst abc) (fst (snd abc)) (snd (snd abc)).
    Proof.
      intros; repeat rewrite <- associativity; apply f_equal; apply functional_extensionality.
      intros; repeat rewrite <- associativity; apply f_equal; apply functional_extensionality.
      intros; repeat rewrite <- associativity, <- left_unit; apply f_equal;
        apply functional_extensionality; intros; rewrite <- left_unit; auto.
    Qed.

    Section PSoundness.

      Variable P_bind : Set.
      Variable E' : Set -> Set.
      Context {Fun_E' : Functor E'}.
      Context {Sub_Except_E' : ExceptE :<: E'}.
      Context {WF_Fun_E' : WF_Functor _ _ Sub_Except_E' _ _}.

      Variable P : UP'_F F -> UP'_F E' -> P_bind -> TypContext -> Prop.

      Variable Catch_Decomp1 :
        forall e1 e2 h1 h2 e1_UP' e2_UP' e1_UP'' e2_UP'' catch1_UP' catch2_UP' pb Sigma,
          P (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _)
              (inj (Catch _ (exist Universal_Property'_fold e2 e2_UP') h2)))) catch2_UP')
          (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _)
              (inj (Catch _ (exist Universal_Property'_fold e1 e1_UP') h1)))) catch1_UP')
          pb Sigma ->
          P (exist _ e2 e2_UP'') (exist _ e1 e1_UP'') pb Sigma.

      Variable Catch_Decomp2 :
        forall e1 e2 h1 h2 catch1_UP' catch2_UP' pb Sigma,
          P (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _)
              (inj (Catch _ e2 h2)))) catch2_UP')
          (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _)
              (inj (Catch _ e1 h1)))) catch1_UP')
          pb Sigma ->
          forall t h1_UP'' h2_UP'',
            P (exist _ (proj1_sig (P := Universal_Property'_fold) (h2 t)) h2_UP'')
            (exist _ (proj1_sig (P := Universal_Property'_fold) (h1 t)) h1_UP'') pb Sigma.

    Global Instance Except_eval_soundness'
      (P_Sub : forall e e' pb Sigma Sigma',
        P e e' pb Sigma -> ConsExtension Sigma' Sigma -> 
        P e e' pb Sigma')
      {Typeof_E' : forall T, FAlgebra TypeofName T (typeofR _ MT) E'}
      {WF_typeof_E' : forall T, @WF_FAlgebra TypeofName T _ _ _
        Sub_Except_E' (MAlgebra_typeof_ExceptE T) (Typeof_E' _)}
      (pb : P_bind)
      (eval_rec : Exp -> evalMR V ME)
      (typeof_rec : UP'_F E' -> typeofR D MT)
      :
      P2Algebra ES'_ExpName ExceptE E' F
      (UP'_P2 (eval_soundness'_P D V F MT ME _ WFVM
        _ _ _ P pb typeof_rec eval_rec (f_algebra (FAlgebra := Typeof_E' _ ))
        (f_algebra (FAlgebra := evalM_F)))).
    Proof.
      apply ind_P2Alg_ExceptE; eauto.
      (* throw case *)
      unfold eval_soundness'_P, UP'_P2; simpl; intros;
        constructor 1 with (x := conj (proj2_sig (inject' (Throw _ n))) (proj2_sig (inject' (Throw _ n)))).
      unfold inject; simpl;
        repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      intros eval_rec_proj typeof_rec_proj Sigma Gamma WF_Sigma_Gamma IHa.
      repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      rewrite (wf_algebra (WF_FAlgebra := WF_eval_F));
        rewrite (wf_algebra (WF_FAlgebra := WF_typeof_E' _));
          simpl fmap; simpl f_algebra; unfold ExceptE_fmap; simpl.
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor.
      (* catch case *)
      intros; destruct e as [eE eT]; destruct (h tt) as [hE hT];
        destruct IHe as [[UP'_eE UP'_eT] IHe];
          destruct (IHh tt) as [[UP'_hE UP'_hT] IHh']; simpl in *|-*.
      unfold eval_soundness'_P, UP'_P2; intros.
      assert (Universal_Property'_fold
        (fst
          (inject
            (Catch (sig Universal_Property'_fold)
              (exist Universal_Property'_fold eE (proj1 (conj UP'_eE UP'_eT)))
              (fun n : unit =>
                exist Universal_Property'_fold (fst (h n))
                (proj1 (proj1_sig (IHh n))))),
            inject
            (Catch (sig Universal_Property'_fold)
              (exist Universal_Property'_fold eT (proj2 (conj UP'_eE UP'_eT)))
              (fun n : unit =>
                exist Universal_Property'_fold (snd (h n))
                (proj2 (proj1_sig (IHh n))))))) /\
        Universal_Property'_fold
        (snd
          (inject
            (Catch (sig Universal_Property'_fold)
              (exist Universal_Property'_fold eE (proj1 (conj UP'_eE UP'_eT)))
              (fun n : unit =>
                exist Universal_Property'_fold (fst (h n))
                (proj1 (proj1_sig (IHh n))))),
            inject
            (Catch (sig Universal_Property'_fold)
              (exist Universal_Property'_fold eT (proj2 (conj UP'_eE UP'_eT)))
              (fun n : unit =>
                exist Universal_Property'_fold (snd (h n))
                (proj2 (proj1_sig (IHh n)))))))) as WF_Catch by
      (simpl; split; unfold inject; exact (proj2_sig _)).
      constructor 1 with (x := WF_Catch); simpl.
      unfold inject; simpl;
        repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      intros eval_rec_proj typeof_rec_proj Sigma WF_Sigma IHa.
      repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      rewrite (wf_algebra (WF_FAlgebra := WF_eval_F));
        rewrite (wf_algebra (WF_FAlgebra := WF_typeof_E' _));
          simpl fmap; simpl f_algebra; unfold ExceptE_fmap.
      intros; apply eval_soundness_H0.
      generalize (IHa _ _ (exist _ _ UP'_eE, exist _ _ UP'_eT)
        (Catch_Decomp1 _ _ _ _ _ _ _ _ _ _ _ _ WF_Sigma)); simpl.
      generalize (typeof_rec_proj (exist _ _ UP'_eE)); simpl;
        intros eE_eq WF_ref; rewrite <- eE_eq; eauto.
      apply WF_ref.
      apply IHe; eauto.
      intros n Sigma' ConsExt_Sig'_Sig; destruct n;
        generalize (IHa _ _ (exist _ _ UP'_hE, exist _ _ UP'_hT)
          (Catch_Decomp2 _ _ _ _ _ _ _ _
            (P_Sub _ _ _ _ _ WF_Sigma ConsExt_Sig'_Sig) tt _ _)); simpl.
      generalize (typeof_rec_proj (exist _ _ UP'_hE)); simpl;
        intros hE_eq WF_ref; rewrite <- hE_eq; eauto.
      apply WF_ref.
      apply IHh'; eauto.
      apply (Catch_Decomp2 _ _ _ _ _ _ _ _
            (P_Sub _ _ _ _ _ WF_Sigma ConsExt_Sig'_Sig) tt _ _).
    Defined.

    End PSoundness.

    Global Instance Except_eval_soundness'' typeof_rec eval_rec :
      P2Algebra ES'_ExpName ExceptE F F
      (UP'_P2 (eval_soundness'_P D V F MT ME _ WFVM
        _ _ _ (fun _ _ _ _ => True) tt typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_F _)) (f_algebra (FAlgebra := evalM_F)))).
    Proof.
      eapply Except_eval_soundness'; eauto.
    Defined.
    
    Section Except_Sound_Sec.

    Variable WFV' : (WFValue_i D V unit -> Prop) -> WFValue_i D V unit -> Prop.
    Variable funWFV' : iFunctor WFV'.
    
    Variable WFVM' : (WFValueM_i D V MT ME unit -> Prop) -> WFValueM_i D V MT ME unit -> Prop.
    Variable funWFVM' : iFunctor WFVM'.

    Definition Except_Sound_P (i : WFValueM_i D V MT ME unit) :=
      forall T, fmap (@proj1_sig _ _) (wfvm_T _ _ _ _ _ i) =
        return_ (proj1_sig T) ->
        (exists v : Value,
          wfvm_v _ _ _ _ _ i = return_ (M := ME) v /\
          WFValueC D V _ WFV' tt v T) \/
        exists t, wfvm_v _ _ _ _ _ i = throw t.

    Inductive Except_Sound_Name := except_sound_name.

    Context {Reasonable_MT : Reasonable_Monad MT}.

    Context {WFV_proj1_a_WFV : iPAlgebra WFV_proj1_a_Name (WFV_proj1_a_P D V _ WFV') WFV'}.
    Context {WFV_proj1_b_WFV : iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V _ WFV') WFV'}.

    Global Instance Except_Sound_WFVM_base :
      iPAlgebra Except_Sound_Name Except_Sound_P (WFValueM_base D V MT ME _ WFV').
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base with (Monad_ME := Mon_M)
        (Fail_MT := Fail_MT) (WFV := WFV');
        try assumption; unfold Except_Sound_P; intros.
      (* WFVM_Return' *)
      left; exists v; split; simpl in *; auto.
      destruct H1 as [mt' mt'_eq]; subst.
      rewrite fmap_fusion in H2.
      destruct (fmap_exists _ _ _ _ _ H2) as [[T' T_eq] T'_eq].
      simpl in *; subst; auto.
      destruct T0 as [T0 T0_UP']; simpl in *.
      destruct Sigma.
      apply (WFV_proj1_b _ _ _ _ _ _ H0); simpl; auto; congruence.
      (* WFVM_Untyped' *)
      left; simpl in H0; apply sym_eq in H0.
      unfold wbind in H0; rewrite fmap_m, <- associativity, bind_fail in H0.
      apply FailMonad_Disc in H0; destruct H0; auto.
    Qed.

    Context {MT_eq_dec : forall (A : Set) (mta : MT A),
      {exists a, mta = return_ a} + {mta = fail}}.
    Context {ME_eq_dec : forall (A : Set) (mte : ME A),
      (exists a, mte = return_ a) \/ (mte = fail) \/ (mte = throw tt)}.
    Context {Inj_MT : InjMonad MT}.
    Context {Reasonable_ME : Reasonable_Monad ME}.
    Context {fail_neq_throw : forall (A B C : Set) (ma : ME A) (mb : ME B),
      (ma >>= fun _ => fail (M := ME)) <> mb >>= fun _ => throw (A := C) tt}.

    Lemma catch_fail : forall (M : Set -> Set) `{FailMonad M} (E : Set) {Exc_M : Exception M E}
      (A B : Set) (f : A -> B) (e : E) (h : E -> M B),
      fail (A := A) = throw (A := A) e ->
        catch (fail (A := B)) h = h e.
    Proof.
      intros.
      assert (fmap f fail (A := A) = fmap f (throw (A := A) e)) by
        (rewrite H0; reflexivity).
      rewrite <- (@fmap_fail _ _ _ _ _ _ f), H1, fmap_throw, catch_throw'; auto.
    Qed.

    Global Instance Except_Sound_WFVM_Except' :
      iPAlgebra Except_Sound_Name Except_Sound_P (WFValueM_Except unit).
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_Except with (TypContextCE := ConsExtensionUnit);
        try assumption; unfold Except_Sound_P; simpl; intros.
      (* throw case *)
      simpl; right; exists tt; reflexivity.
      (* catch case *)
      destruct (MT_eq_dec _ mte) as [[T' mte_eq] | mte_eq]; subst.
      destruct (MT_eq_dec _ mth) as [[T'' mth_eq] | mth_eq]; subst.
      repeat rewrite <- left_unit in H2.
      caseEq (eq_DType _ (proj1_sig T') T''); rewrite H3 in H2.
      destruct (H0 _ H2) as [[v' [ek'_eq WF_v'_T']] | [[] ek'_eq]].
      destruct (ME_eq_dec _ e') as [[v'' e'_eq] | [e'_eq | e'_eq]];
        subst.
      rewrite catch_return; left; exists v'; split; auto.
      rewrite bind_fail in ek'_eq.
      elimtype False; eapply FailMonad_Disc with (M := ME) (a := v') (mb := return_ v');
        eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; auto.
      rewrite bind_throw in ek'_eq.
      elimtype False; eapply Exception_Disc with (M := ME) (a := v') (e := tt) (mb := return_ v');
        eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; auto.
      destruct (ME_eq_dec _ e') as [[v'' e'_eq] | [e'_eq | e'_eq]];
        subst.
      rewrite catch_return; right; eexists _; eauto.
      rewrite bind_fail in ek'_eq; elimtype False;
      rewrite left_unit with (a := tt) (f := fun _ => throw tt) in ek'_eq;
        rewrite left_unit with (a := tt) (f := fun _ => fail) in ek'_eq;
          eapply fail_neq_throw; eauto.
      rewrite catch_throw'.
      destruct (H1 tt Sigma (ConsExtension_id _) T) as [[v' [h_eq WF_v'_T']] | [[] h_eq]].
      rewrite rewrite_do with (m := fun U => return_ T' >>= kT U)
        (m' := fun U => kT U T') in H2; [auto |
          apply functional_extensionality; intros; repeat rewrite <- left_unit; auto].
      rewrite rewrite_do with (m := fun U => return_ T'' >>= kT U)
        (m' := fun U => kT U T''); [auto |
          apply functional_extensionality; intros; repeat rewrite <- left_unit; auto].
      rewrite <- H2; apply kt_eq; apply sym_eq; apply (eq_DType_eq _ _ _ H3).
      left; exists v'; split; auto.
      right; eauto.
      destruct (MT_eq_dec _ kT') as [[U kT'_eq] | kT'_eq]; subst.
      rewrite <- left_unit, bind_fail, fmap_fail in H2.
      elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := return_ (M := MT) (proj1_sig T));
        eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; auto.
      rewrite bind_fail, fmap_fail in H2.
      elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := return_ (M := MT) (proj1_sig T));
        eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; auto.
      destruct (MT_eq_dec _ kT') as [[U kT'_eq] | kT'_eq]; subst.
      rewrite <- left_unit, <- left_unit, bind_fail, bind_fail, fmap_fail in H2.
      elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := return_ (M := MT) (proj1_sig T));
        eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; auto.
      rewrite bind_fail, fmap_fail in H2.
      elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := return_ (M := MT) (proj1_sig T));
        eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; auto.
      destruct (MT_eq_dec _ kT') as [[U kT'_eq] | kT'_eq]; subst.
      rewrite <- left_unit, bind_fail, bind_fail, fmap_fail in H2.
      elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := return_ (M := MT) (proj1_sig T));
        eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; auto.
      rewrite bind_fail, fmap_fail in H2.
      elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := return_ (M := MT) (proj1_sig T));
        eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; auto.
    Qed.

    Context {Except_Sound_WFVM : iPAlgebra Except_Sound_Name Except_Sound_P WFVM'}.

    Context {eval_soundness'_Exp_F : forall (typeof_rec : UP'_F F -> typeofR D MT)
       (eval_rec : Names.Exp F -> evalMR V ME),
       P2Algebra ES'_ExpName F F F
       (UP'_P2
         (eval_soundness'_P D V F MT ME _ WFVM'
           Datatypes.unit F Fun_F
           (fun _ _ _ _ => True)
           tt typeof_rec eval_rec f_algebra
           (f_algebra (FAlgebra := evalM_F' (@Names.Exp F Fun_F)))))}.

     Context {WF_MAlg_typeof : WF_MAlgebra Typeof_F}.
     Context {WF_MAlg_eval : WF_MAlgebra evalM_F'}.

    Theorem eval_Except_Sound :
      forall (e : Exp) (T : DType D),
        fmap (@proj1_sig _ _) (typeof D F MT (proj1_sig e)) = return_ (proj1_sig T) ->
        (exists v : Value,
          evalM V F ME (proj1_sig e) = return_ (M := ME) v /\
          WFValueC D V _ WFV' tt v T) \/
        exists t, evalM V F ME (proj1_sig e) = throw (M := ME) t.
    Proof.
      intro e.
      apply (ifold_ WFVM' _ (ip_algebra (iPAlgebra := Except_Sound_WFVM)) _
        (eval_soundness (eval_soundness'_Exp_E := eval_soundness'_Exp_F) D V F MT ME WFVM' e tt)).
    Qed.

  End Except_Sound_Sec.

    (* Lemma bevalM_continuous_Exp'_H : forall n,
      UP'_SP (bevalM_continuous_Exp'_P V F ME SV) (throw' n).
      unfold bevalM_continuous_Exp'_P; intros; econstructor; intros.
      auto with typeclass_instances.
      unfold Names.bevalM, mfold; simpl; rewrite H0; unfold throw'.
      WF_FAlg_rewrite.
      assert (SubValueC _ SV (bot' V) (bot' V)) by (apply inject_i; econstructor; eauto).
      exists (return_ (exist (fun i => SubValue V SV i) _ H1), throw n); split; simpl; rewrite fmap_m.
      rewrite <- left_unit; simpl; auto.
      rewrite bind_throw; auto.
    Qed.

  Lemma bevalM_continuous_Exp'_H0 : forall (e : Fix F) (h : nat -> Fix F)
      (IHe : UP'_SP (bevalM_continuous_Exp'_P V F ME SV) e)
      (IHh : forall n, UP'_SP (bevalM_continuous_Exp'_P V F ME SV) (h n)),
      UP'_SP (bevalM_continuous_Exp'_P V F ME SV)
      (@catch' _ _ (projT1 IHe) (fun n => projT1 (IHh n))).
      unfold bevalM_continuous_Exp'_P; intros.
      econstructor; intros; eauto with typeclass_instances.
      unfold Names.bevalM, mfold, catch'; simpl; rewrite H0; unfold catch'; simpl.
      WF_FAlg_rewrite.
      destruct IHe as [e_UP' IHe].
      repeat erewrite bF_UP_in_out.
      assert (SubValueC _ SV (bot' V) (bot' V)) by (apply inject_i; econstructor; eauto).
      exists (catch (fst (projT1 (H (exist _ _ e_UP') (exist _ _ e_UP') (refl_equal _))))
                    (fun e => fst (projT1 (H (exist _ (h e) (projT1 (IHh e)))
                      (exist _ (h e) (projT1 (IHh e))) (refl_equal _)))),
              catch (snd (projT1 (H (exist _ _ e_UP') (exist _ _ e_UP') (refl_equal _))))
                    (fun e => snd (projT1 (H (exist _ (h e) (projT1 (IHh e)))
                      (exist _ (h e) (projT1 (IHh e))) (refl_equal _))))); simpl.
      repeat rewrite fmap_catch.
      repeat rewrite fmap_m; split; simpl.
      destruct ((H (exist Universal_Property'_fold e e_UP')
                                (exist Universal_Property'_fold e e_UP')
                                (eq_refl e))) as [[Mc Mh] [Mc_eq Mh_eq']]; simpl in *|-*.
      rewrite fmap_m in Mc_eq; rewrite Mc_eq; rewrite catch_return; auto.
      apply f_equal2.
      destruct ((H (exist Universal_Property'_fold e e_UP')
                                (exist Universal_Property'_fold e e_UP')
                                (eq_refl e))) as [[Mc Mh] [Mc_eq Mh_eq']]; simpl in *|-*.
      rewrite <- fmap_m.
      rewrite Mh_eq'; auto.
      apply functional_extensionality; intro.
      destruct (projT2 (H (exist _ (h x) (projT1 (IHh x))) (exist _ (h x) (projT1 (IHh x)))
        (refl_equal _))) as [Mh_eq Mh_eq'].
      rewrite <- fmap_ProductM_snd in Mh_eq'; auto; simpl in Mh_eq'.
      erewrite Mh_eq'.
      erewrite bF_UP_in_out; reflexivity.
    Qed.

    Global Instance Exception_bevalM_continuous_Exp' :
      DAlgebra EC_ExpName Except F (UP'_SP (bevalM_continuous_Exp'_P V F ME SV)).
    Proof.
      apply ind_DAlg_Except.
      apply bevalM_continuous_Exp'_H.
      apply bevalM_continuous_Exp'_H0.
    Qed.

    (* Lemma WF_ind_alg_Arith (Name : Set)
    (P : forall e : Fix F, Universal_Property'_fold e -> Prop)
    (H : forall n, UP'_P P (lit n))
    (H0 : forall m n
      (IHm : UP'_P P m)
      (IHn : UP'_P P n),
      UP'_P P (@add m n (proj1_sig IHm) (proj1_sig IHn)))
    {Sub_Arith_F' : Arith :<: F} :
    (forall a, inj (Sub_Functor := Sub_Arith_F) a =
      inj (A := (Fix F)) (Sub_Functor := Sub_Arith_F') a) ->
      WF_Ind (Name := Name) {| p_algebra := ind_alg_Arith P H H0|}.
      constructor; intros.
      simpl; unfold ind_alg_Arith; destruct e; simpl.
      unfold lit; simpl; rewrite wf_functor; simpl; apply f_equal; eauto.
      unfold add; simpl; rewrite wf_functor; simpl; apply f_equal; eauto.
    Defined. *)

    Context {eval_continuous_Exp_E :
      PAlgebra EC_ExpName F F (UP'_P (eval_continuous_Exp_P V F SV))}.

  (* ============================================== *)
  (* WELL-FORMED NAT VALUES                         *)
  (* ============================================== *)

    Variable WFV : (WFValue_i D V -> Prop) -> WFValue_i D V -> Prop.
    Variable funWFV : iFunctor WFV.

    (** Natrual Numbers are well-formed **)

    Inductive WFValue_VI (WFV : WFValue_i D V -> Prop) : WFValue_i D V -> Prop :=
    | WFV_VI : forall n v T,
      proj1_sig v = vi n ->
      proj1_sig T = tnat ->
      WFValue_VI WFV (mk_WFValue_i D V v T).

    Definition ind_alg_WFV_VI (P : WFValue_i D V -> Prop)
      (H : forall n v T veq Teq, P (mk_WFValue_i _ _ v T))
      i (e : WFValue_VI P i) : P i :=
      match e in WFValue_VI _ i return P i with
        | WFV_VI n v T veq Teq => H n v T veq Teq
      end.

    Definition WFV_VI_ifmap (A B : WFValue_i D V -> Prop) i (f : forall i, A i -> B i)
      (WFV_a : WFValue_VI A i) : WFValue_VI B i :=
      match WFV_a in (WFValue_VI _ s) return (WFValue_VI B s)
        with
        | WFV_VI n v T veq Teq => WFV_VI B n v T veq Teq
      end.

    Global Instance iFun_WFV_VI : iFunctor WFValue_VI.
      constructor 1 with (ifmap := WFV_VI_ifmap).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; reflexivity.
    Defined.

    Variable Sub_WFV_VI_WFV : Sub_iFunctor WFValue_VI WFV.

     Global Instance WFV_proj1_a_VI :
       iPAlgebra WFV_proj1_a_Name (WFV_proj1_a_P D V WFV) WFValue_VI.
       econstructor; intros.
       unfold iAlgebra; intros; unfold WFV_proj1_a_P.
       inversion H; subst; simpl; intros.
       apply (inject_i (subGF := Sub_WFV_VI_WFV)); econstructor; simpl; eauto.
       rewrite H3; eauto.
     Defined.

     Global Instance WFV_proj1_b_VI :
       iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V WFV) WFValue_VI.
       econstructor; intros.
       unfold iAlgebra; intros; unfold WFV_proj1_b_P.
       inversion H; subst; simpl; intros.
       apply (inject_i (subGF := Sub_WFV_VI_WFV)); econstructor; simpl; eauto.
       rewrite H3; eauto.
     Defined.

    (* Inversion principles for Well-formed natural numbers. *)
    Definition WF_invertVI_P (i : WFValue_i D V) :=
      proj1_sig (wfv_b _ _ i) = tnat ->
      WFValue_VI (iFix WFV) i \/ (proj1_sig (wfv_a D V i) = bot V).

    Inductive WF_invertVI_Name := wfv_invertvi_name.
    Context {WF_invertVI_WFV :
      iPAlgebra WF_invertVI_Name WF_invertVI_P WFV}.

    Global Instance WF_invertVI_VI :
      iPAlgebra WF_invertVI_Name WF_invertVI_P WFValue_VI.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WF_invertVI_P.
      inversion H; subst; simpl; intros.
      left; econstructor; eassumption.
    Defined.

    Global Instance WF_invertVI_Bot :
      iPAlgebra WF_invertVI_Name WF_invertVI_P (WFValue_Bot _ _).
      econstructor; intros.
      unfold iAlgebra; intros; unfold WF_invertVI_P.
      inversion H; subst; simpl; intros.
      inversion H; subst; rewrite H3; right; reflexivity.
    Defined.

    Definition WF_invertVI := ifold_ WFV _ (ip_algebra (iPAlgebra := WF_invertVI_WFV)).

    Context {WFV_proj1_a_WFV :
      iPAlgebra WFV_proj1_a_Name (WFV_proj1_a_P D V WFV) WFV}.
    Context {WFV_proj1_b_WFV :
       iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V WFV) WFV}.

    Lemma Arith_eval_Soundness_H
      (typeof_R eval_R : Set) typeof_rec eval_rec
      {eval_F' : FAlgebra EvalName eval_R (evalR V) F}
      {WF_eval_F' : @WF_FAlgebra EvalName _ _ Arith F
        Sub_Arith_F (MAlgebra_eval_Arith _) (eval_F')} :
      forall n : nat,
        forall gamma'' : Env (Names.Value V),
          forall T : Names.DType D,
            Arith_typeof typeof_R typeof_rec (Lit _ n) = Some T ->
            WFValueC D V WFV (Arith_eval eval_R eval_rec (Lit _ n) gamma'') T.
      intros n gamma'' T H4; intros.
      apply (inject_i (subGF := Sub_WFV_VI_WFV)); econstructor; eauto.
      simpl.
      unfold vi, vi', inject; simpl; eauto.
      unfold typeof, mfold, lit in H4; simpl in H4.
      injection H4; intros; subst.
      reflexivity.
    Defined.

   Lemma Arith_eval_Soundness_H0
     (typeof_R eval_R : Set) typeof_rec eval_rec
     {eval_F' : FAlgebra EvalName eval_R (evalR V) F}
     {WF_eval_F' : @WF_FAlgebra EvalName _ _ Arith F
        Sub_Arith_F (MAlgebra_eval_Arith _) (eval_F')} :
      forall (a b : typeof_R) (a' b' : eval_R),
        forall gamma'' : Env (Names.Value V),
          (forall T : Names.DType D,
            typeof_rec a = Some T ->
            WFValueC D V WFV (eval_rec a' gamma'') T) ->
          (forall T : Names.DType D,
            typeof_rec b = Some T ->
            WFValueC D V WFV (eval_rec b' gamma'') T) ->
          forall T : Names.DType D,
            Arith_typeof typeof_R typeof_rec (Add _ a b) = Some T ->
            WFValueC D V WFV (Arith_eval eval_R eval_rec (Add _ a' b') gamma'') T.
     simpl; intros a b a' b' gamma'' IH_a IH_b T H4.
      caseEq (typeof_rec a); intros; rename H into typeof_a;
        unfold typeofR in typeof_a, H4; rewrite typeof_a in H4;
          try discriminate.
      caseEq (typeof_rec b); intros; rename H into typeof_b;
        unfold typeofR in typeof_b, H4; rewrite typeof_b in H4;
          try discriminate.
      caseEq (isTNat (proj1_sig d)); intros; rename H into d_eq; rewrite
        d_eq in H4; try discriminate.
      caseEq (isTNat (proj1_sig d0)); intros; rename H into d0_eq; rewrite
        d0_eq in H4; try discriminate.
      injection H4; intros; subst; clear H4.
      unfold isTNat in d_eq, d0_eq.
      caseEq (project (proj1_sig d)); intros; rewrite H in d_eq;
        try discriminate; clear d_eq; rename H into d_eq; destruct a0.
      caseEq (project (proj1_sig d0)); intros; rewrite H in d0_eq;
        try discriminate; clear d0_eq; rename H into d0_eq; destruct a0.
      apply project_inject in d_eq; apply project_inject in d0_eq;
        eauto with typeclass_instances.
      generalize (IH_a _ typeof_a) as WF_a;
      generalize (IH_b _ typeof_b) as WF_b; intros.
      unfold WFValueC in *|-*.
      destruct (WF_invertVI _ WF_a d_eq) as [beval_a' | beval_a'];
        inversion beval_a'; subst.
      rewrite H1; unfold isVI, project, vi, vi', inject'; simpl;
        rewrite out_in_fmap; repeat rewrite wf_functor; simpl; rewrite prj_inj.
      destruct (WF_invertVI _ WF_b d0_eq) as [beval_b' | beval_b'];
        inversion beval_b'; subst.
      rewrite H3; unfold isVI, project, vi, vi', inject'; simpl;
        rewrite out_in_fmap; repeat rewrite wf_functor; simpl; rewrite prj_inj.
      apply (inject_i (subGF := Sub_WFV_VI_WFV)); econstructor; eauto.
      simpl; rewrite wf_functor; simpl; eauto.
      unfold vi, vi', inject'; simpl; eauto;
        rewrite wf_functor; simpl; reflexivity.
      rewrite H0; unfold bot, isVI, project, inject, inject'; simpl;
        rewrite out_in_fmap; repeat rewrite wf_functor; simpl; unfold Bot_fmap.
      caseEq (prj (Sub_Functor := Sub_NatValue_V) (A:= (sig (@Universal_Property'_fold V _)))
        (inj (Bot (sig Universal_Property'_fold)))).
      elimtype False; eapply (inject_discriminate Dis_VI_Bot n0).
      unfold inject; simpl; apply inj_prj in H; erewrite <- H; reflexivity.
      caseEq (isBot V (in_t (inj (VI (Fix V) n)))).
      simpl in beval_b'.
      apply sym_eq in H0; apply sym_eq in d0_eq.
      generalize (WFV_proj1_a D V WFV _ _ WF_b _ _ H0).
      simpl; intros WF_b'; generalize (WFV_proj1_b D V WFV _ _ WF_b' _ _ d0_eq); simpl.
      assert (exist Universal_Property'_fold (bot V) (UP'_bot V) = bot' V) by
        (unfold bot, UP'_bot; destruct bot'; reflexivity).
      assert (exist Universal_Property'_fold tnat UP'_tnat = tnat') by
        (unfold tnat, UP'_tnat; destruct tnat'; reflexivity).
      unfold tnat, tnat', inject' at 1 in H5.
      unfold inject.
      rewrite H5, H4; auto.
      unfold isBot, project; simpl; rewrite out_in_fmap;
        repeat rewrite wf_functor; simpl; unfold Bot_fmap;
          rewrite prj_inj.
      simpl in beval_b'.
      apply sym_eq in H0; apply sym_eq in d0_eq.
      unfold eval in WF_b; generalize (WFV_proj1_a D V WFV _ _ WF_b _ _ H0).
      simpl; intros WF_b'; generalize (WFV_proj1_b D V WFV _ _ WF_b' _ _ d0_eq); simpl.
      assert (exist Universal_Property'_fold (bot V) (UP'_bot V) = bot' V) by
        (unfold bot, UP'_bot; destruct bot'; reflexivity).
      assert (exist Universal_Property'_fold tnat UP'_tnat = tnat') by
        (unfold tnat, UP'_tnat; destruct tnat'; reflexivity).
      unfold tnat, tnat', inject' at 1 in H5.
      unfold inject.
      rewrite H5, H4; auto.
      caseEq (isVI (proj1_sig (eval_rec a' gamma''))).
      unfold isVI in H; rewrite H0 in H.
      unfold bot, isVI, project, inject, inject' in H; simpl in H;
        rewrite out_in_fmap in H; repeat rewrite wf_functor in H; simpl in H;
          unfold Bot_fmap in H.
      caseEq (prj (Sub_Functor := Sub_NatValue_V) (A:= (sig (@Universal_Property'_fold V _)))
        (inj (Bot (sig Universal_Property'_fold)))).
      elimtype False; eapply (inject_discriminate Dis_VI_Bot n0);
        apply inj_prj in H1; unfold inject; rewrite <- H1; reflexivity.
      rewrite H1 in H; discriminate.
      rewrite H0.
      unfold bot, isBot, project, inject; simpl; rewrite out_in_fmap;
        repeat rewrite wf_functor; simpl; unfold Bot_fmap; simpl.
      simpl in beval_a'.
      apply sym_eq in H0; apply sym_eq in d_eq.
      generalize (WFV_proj1_a D V WFV _ _ WF_a _ _ H0).
      simpl; intros WF_a'; generalize (WFV_proj1_b D V WFV _ _ WF_a' _ _ d_eq); simpl.
      assert (exist Universal_Property'_fold (bot V) (UP'_bot V) = bot' V) by
        (unfold bot, UP'_bot; destruct bot'; reflexivity).
      assert (exist Universal_Property'_fold tnat UP'_tnat = tnat') by
        (unfold tnat, UP'_tnat; destruct tnat'; reflexivity).
      unfold tnat, tnat', inject' at 1 in H2.
      unfold inject.
      rewrite H2, H1, prj_inj; auto.
      exact (proj2_sig d0).
      exact (proj2_sig d0).
      exact (proj2_sig d).
      exact (proj2_sig d0).
    Defined.

    Context {Typeof_F : forall T, FAlgebra TypeofName T (typeofR _) F}.
    Context {WF_typeof_F : forall T, @WF_FAlgebra TypeofName T _ _ _
      Sub_Arith_F (MAlgebra_typeof_Arith T) (Typeof_F _)}.
    Context {WF_Value_continous_alg :
      iPAlgebra WFV_ContinuousName (WF_Value_continuous_P D V WFV) SV}.

    Global Instance Arith_eval_Soundness_alg
      (P_bind : Set)
      (P : P_bind -> Env Value -> Prop)
      (E' : Set -> Set)
      {Fun_E' : Functor E'}
      {Sub_Arith_E' : Arith :<: E'}
      {WF_Fun_E' : WF_Functor _ _ Sub_Arith_E' _ _}
      {Typeof_E' : forall T, FAlgebra TypeofName T (typeofR _) E'}
      {WF_typeof_E' : forall T, @WF_FAlgebra TypeofName T _ _ _
        Sub_Arith_E' (MAlgebra_typeof_Arith T) (Typeof_E' _)}
      (pb : P_bind)
      (eval_rec : Exp -> evalR V)
      (typeof_rec : UP'_F E' -> typeofR D)
      :
      P2Algebra eval_Soundness_alg_Name Arith E' F
      (UP'_P2 (@eval_alg_Soundness_P D _ V _ _ _ WFV _ P
        _ _ pb typeof_rec eval_rec (f_algebra (FAlgebra := Typeof_E' _ ))
        (f_algebra (FAlgebra := eval_F)))).
    Proof.
      apply ind_P2Alg_Arith; try eassumption.
      unfold eval_alg_Soundness_P, UP'_P2; intros.
      constructor.
      exact (conj (proj2_sig (inject' (Lit _ n))) (proj2_sig (lit' n))).
      unfold lit, lit', inject; simpl.
      repeat rewrite out_in_fmap.
      repeat rewrite wf_functor.
      intros gamma'' WF_gamma'' IHa.
      rewrite (wf_algebra (WF_FAlgebra := WF_eval_F));
        rewrite (wf_algebra (WF_FAlgebra := WF_typeof_E' _));
          simpl fmap; simpl f_algebra; unfold Arith_fmap.
      intros.
      eapply Arith_eval_Soundness_H.
      apply WF_eval_F.
      apply H.
      (* Add Case *)
      unfold eval_alg_Soundness_P, UP'_P2; intros.
      destruct m as [m1 m2]; destruct n as [n1 n2];
        destruct IHm as [[UP'_m1 UP'_m2] IHm];
          destruct IHn as [[UP'_n1 UP'_n2] IHn]; simpl in *|-*.
      constructor.
      split; unfold inject; exact (proj2_sig _).
      unfold inject; simpl;
        repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      intros eval_rec_proj typeof_rec_proj gamma'' WF_gamma'' IHa.
      rewrite (wf_algebra (WF_FAlgebra := WF_eval_F));
        rewrite (wf_algebra (WF_FAlgebra := WF_typeof_E' _));
          simpl fmap; simpl f_algebra; unfold Arith_fmap.
      intros T; eapply Arith_eval_Soundness_H0.
      apply WF_eval_F.
      apply (IHa _ _ WF_gamma'' (_, exist _ _ UP'_m2)); simpl;
        intros; apply (IHm eval_rec_proj typeof_rec_proj _ WF_gamma'' IHa);
          intros; auto; rewrite <- (in_out_UP'_inverse _ _ m1); auto.
      apply (IHa _ _ WF_gamma'' (_, exist _ _ UP'_n2)); simpl;
        intros; apply (IHn eval_rec_proj typeof_rec_proj _ WF_gamma'' IHa);
          intros; auto; rewrite <- (in_out_UP'_inverse _ _ n1); auto.
    Defined.

End Arith.

  Hint Extern 1 (iPAlgebra SV_invertVI_Name (SV_invertVI_P _) _) =>
      constructor; unfold iAlgebra; unfold SV_invertVI_P; intros i H n H0;
        inversion H; subst; simpl in H0; revert H0;
          match goal with H : proj1_sig ?v = _ |- proj1_sig ?v = _ -> _ => rewrite H end; intros;
            elimtype False; apply (inject_discriminate _ _ _ H0).

  Hint Extern 1 (iPAlgebra SV_invertVI'_Name (SV_invertVI'_P _) _) =>
      constructor; unfold iAlgebra; unfold SV_invertVI'_P; intros i H n H0;
        inversion H; subst; simpl in H0; revert H0;
          match goal with H : proj1_sig ?v = _ |- proj1_sig ?v = _ -> _ => rewrite H end; intros;
            elimtype False; apply (inject_discriminate _ _ _ H0).

  Hint Extern 5 (iPAlgebra WF_invertVI_Name (WF_invertVI_P _ _ _) _) =>
    constructor; unfold iAlgebra; let H := fresh in
      intros i H; unfold WF_invertVI_P;
        inversion H; simpl;
          match goal with
            eq_H0 : proj1_sig ?T = _ |- proj1_sig ?T = _ -> _ =>
              intro eq_H; rewrite eq_H in eq_H0;
                elimtype False; first [apply (inject_discriminate _ _ _ eq_H0) |
                  apply sym_eq in eq_H0; apply (inject_discriminate _ _ _ eq_H0)]
          end : typeclass_instances. *)

  End Exception.

(* Hint Extern 0 =>
  intros; match goal with
            |- (PAlgebra eval_Soundness_alg_Name
              (sig (UP'_P2 (eval_alg_Soundness_P _ _ _ _ _ _ _ _ _ _ _ _ _ _))) Arith) =>
            eapply Arith_eval_Soundness_alg; eauto with typeclass_instances
          end : typeclass_instances. *)


(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
