Require Import FJ_tactics.
Require Import Functors.
Require Import List.
Require Import Names.
Require Import FunctionalExtensionality.
Require Import Coq.Arith.EqNat.
Require Import Coq.Bool.Bool.
Require Import MonadLib.

Section PNames.

  (* ============================================== *)
  (* TYPES                                          *)
  (* ============================================== *)

  (** SuperFunctor for Types. **)
  Variable D : Set -> Set.
  Context {Fun_D : Functor D}.

  (* ============================================== *)
  (* VALUES                                         *)
  (* ============================================== *)

  (** SuperFunctor for Values. **)
  Variable V : Set -> Set.
  Context {Fun_V : Functor V}.

  Context {Sub_StuckValue_V : StuckValue :<: V}.
  Context {WF_SubStuckValue_V : WF_Functor _ _ Sub_StuckValue_V _ _}.

  (* ============================================== *)
  (* EXPRESSIONS                                    *)
  (* ============================================== *)

  (** SuperFunctor for Expressions. **)
  Variable E : Set -> Set -> Set.
  Context {Fun_E : forall A, Functor (E A)}.

  (* ============================================== *)
  (* OPERATIONS                                     *)
  (* ============================================== *)

  (** TYPING **)

  Definition MT := Maybe.

  Context {Typeof_E : forall T,
    FAlgebra TypeofName T (typeofR D MT) (E (DType D))}.

  Variable ME : Set -> Set.
  Context {Fun_ME : Functor ME}.
  Context {Mon_ME : Monad ME}.
  Context {Fail_ME : FailMonad ME}.

  Context {bevalM_E : FAlgebra EvalName (Exp (E nat)) (evalMR V ME) (E nat)}.

  Definition bevalM (n: nat) := boundedFix_UP n (@f_algebra _ _ _ _ bevalM_E) (fail (A := Value V)).

  Section WFValue_Sec.

  Variable TypContext : Set.
  Context {TypContextCE : ConsExtensionC TypContext}.

  (* ============================================== *)
  (* WELL-FORMED NONTERMINATION                     *)
  (* ============================================== *)

  Inductive WFValueM_Bot (A : WFValueM_i D V MT ME TypContext -> Prop) :
    WFValueM_i D V MT ME TypContext -> Prop :=
  | WFVM_fail : forall Sigma T,
    WFValueM_Bot A (mk_WFValueM_i D V MT ME TypContext Sigma fail T).
  
  Definition ind_alg_WFVM_Bot (P : WFValueM_i D V MT ME _ -> Prop)
    (H_fail : forall Sigma T,
      P (mk_WFValueM_i _ _ _ _ _ Sigma fail T))
      (i : WFValueM_i D V MT ME _)
      (e : WFValueM_Bot P i)
      :
      P i :=
      match e in WFValueM_Bot _ i return P i with
        | WFVM_fail Sigma T => H_fail Sigma T
      end.

  Definition WFVM_Bot_ifmap
    (A B : WFValueM_i D V MT ME _ -> Prop)
    (i : WFValueM_i D V MT ME _)
    (f : forall i, A i -> B i)
    (WFVM_a : WFValueM_Bot A i)
    :
    WFValueM_Bot B i :=
    match WFVM_a in WFValueM_Bot _ i return WFValueM_Bot B i with
      | WFVM_fail Sigma T => WFVM_fail B Sigma T
    end.

    Global Instance iFun_wfvm_Bot : iFunctor WFValueM_Bot.
      constructor 1 with (ifmap := WFVM_Bot_ifmap).
      (* ifmap_fusion *)
      destruct a; simpl; intros; reflexivity.
      (* ifmap_id *)
      destruct a; simpl; intros; auto;
        apply f_equal; repeat (apply functional_extensionality_dep; intros); auto.
    Defined.

  (* ============================================== *)
  (* EXPRESSION EQUIVALENCE RELATION                *)
  (* ============================================== *)

  Section eqv_Section.

    Record eqv_i (A B : Set) : Set := mk_eqv_i
      {env_A : Env A;
        env_B : Env B;
        eqv_a : UP'_F (E A);
        eqv_b : UP'_F (E B)}.

    (** SuperFunctor for Equivalence Relation. **)
    Variable EQV_E : forall A B,
      (@eqv_i A B -> Prop) -> eqv_i A B -> Prop.

    Definition E_eqv A B := iFix (EQV_E A B).
    Definition E_eqvC {A B : Set} gamma gamma' e e' :=
      E_eqv _ _ (mk_eqv_i A B gamma gamma' e e').

    Variable (NP : Set -> Set).
    Context {Fun_NP : Functor NP}.
    Context {sub_NP_F : forall A, NP :<: E A}.

    Inductive NP_Functor_eqv
      (A B : Set)
      (C : eqv_i A B -> Prop)
      : eqv_i A B -> Prop :=
    | NP_eqv_0 : forall (gamma : Env A) (gamma' : Env B)
      e e' (np : forall D : Set, NP D),
      proj1_sig e = inject (subGF := sub_NP_F A) (np _) ->
      proj1_sig e' = inject (subGF := sub_NP_F B) (np _) ->
      (forall (A' B' C' : Set) (f : A' -> C') (g : B' -> C'),
        fmap f (np A') = fmap g (np B')) ->
      NP_Functor_eqv A B C (mk_eqv_i _ _ gamma gamma' e e')
    | NP_eqv_1 : forall (gamma : Env A) (gamma' : Env B)
      e e' (np : forall D : Set, D -> NP D)
      a a',
      proj1_sig e = inject (subGF := sub_NP_F A) (np _ a) ->
      proj1_sig e' = inject (subGF := sub_NP_F B) (np _ a') ->
      C (mk_eqv_i _ _ gamma gamma' a a') ->
      (forall (A' B' C' : Set) (f : A' -> C') (g : B' -> C')
        (a' : A') (b' : B'), f a' = g b' ->
        fmap f (np A' a') = fmap g (np B' b')) ->
      NP_Functor_eqv A B C (mk_eqv_i _ _ gamma gamma' e e')
    | NP_eqv_2 : forall (gamma : Env A) (gamma' : Env B)
      e e' (np : forall D : Set, D -> D -> NP D)
      a a' b b',
      proj1_sig e = inject (subGF := sub_NP_F A) (np _ a b) ->
      proj1_sig e' = inject (subGF := sub_NP_F B) (np _ a' b') ->
      C (mk_eqv_i _ _ gamma gamma' a a') ->
      C (mk_eqv_i _ _ gamma gamma' b b') ->
      (forall (A' B' C' : Set) (f : A' -> C') (g : B' -> C')
        (a' a'' : A') (b' b'' : B'), f a' = g b' -> f a'' = g b'' ->
        fmap f (np A' a' a'') = fmap g (np B' b' b'')) ->
      NP_Functor_eqv A B C (mk_eqv_i _ _ gamma gamma' e e')
    | NP_eqv_3 : forall (gamma : Env A) (gamma' : Env B)
      e e' (np : forall D : Set, D -> D -> D -> NP D)
      a a' b b' c c',
      proj1_sig e = inject (subGF := sub_NP_F A) (np _ a b c) ->
      proj1_sig e' = inject (subGF := sub_NP_F B) (np _ a' b' c') ->
      C (mk_eqv_i _ _ gamma gamma' a a') ->
      C (mk_eqv_i _ _ gamma gamma' b b') ->
      C (mk_eqv_i _ _ gamma gamma' c c') ->
      (forall (A' B' C' : Set) (f : A' -> C') (g : B' -> C')
        (a' a'' a''' : A') (b' b'' b''' : B'),
        f a' = g b' -> f a'' = g b'' -> f a''' = g b''' ->
        fmap f (np A' a' a'' a''') = fmap g (np B' b' b'' b''')) ->
      NP_Functor_eqv A B C (mk_eqv_i _ _ gamma gamma' e e').

    Definition ind_alg_NP_Functor_eqv
      (A B : Set)
      (P : eqv_i A B -> Prop)
      (H : forall gamma gamma' e e' np e_eq e'_eq wf_np,
        P (mk_eqv_i _ _ gamma gamma' e e'))
      (H0 : forall gamma gamma' e e' np a a' e_eq e'_eq
        (IHa : P (mk_eqv_i _ _  gamma gamma' a a')) wf_np,
        P (mk_eqv_i _ _  gamma gamma' e e'))
      (H1 : forall gamma gamma' e e' np a a' b b' e_eq e'_eq
        (IHa : P (mk_eqv_i _ _  gamma gamma' a a'))
        (IHb : P (mk_eqv_i _ _  gamma gamma' b b')) wf_np,
        P (mk_eqv_i _ _  gamma gamma' e e'))
      (H2 : forall gamma gamma' e e' np a a' b b' c c' e_eq e'_eq
        (IHa : P (mk_eqv_i _ _  gamma gamma' a a'))
        (IHb : P (mk_eqv_i _ _  gamma gamma' b b'))
        (IHc : P (mk_eqv_i _ _  gamma gamma' c c')) wf_np,
        P (mk_eqv_i _ _  gamma gamma' e e'))
      i (e : NP_Functor_eqv A B P i) : P i :=
      match e in NP_Functor_eqv _ _ _  i return P i with
        | NP_eqv_0 gamma gamma' e e' np e_eq e'_eq wf_np =>
          H gamma gamma' e e' np e_eq e'_eq wf_np
        | NP_eqv_1 gamma gamma' e e' np a a' e_eq e'_eq a_eqv wf_np =>
          H0 gamma gamma' e e' np a a' e_eq e'_eq a_eqv wf_np
        | NP_eqv_2 gamma gamma' e e' np a a' b b' e_eq e'_eq a_eqv b_eqv wf_np =>
          H1 gamma gamma' e e' np a a' b b' e_eq e'_eq a_eqv b_eqv wf_np
        | NP_eqv_3 gamma gamma' e e' np a a' b b' c c' e_eq e'_eq a_eqv b_eqv c_eqv wf_np  =>
          H2 gamma gamma' e e' np a a' b b' c c' e_eq e'_eq a_eqv b_eqv c_eqv wf_np
      end.

    Definition NP_Functor_eqv_ifmap
      (A B : Set)
      (A' B' : eqv_i A B -> Prop) i (f : forall i, A' i -> B' i)
      (eqv_a : NP_Functor_eqv A B A' i) :
      NP_Functor_eqv A B B' i :=
      match eqv_a in NP_Functor_eqv _ _ _ i return NP_Functor_eqv _ _ _ i with
        | NP_eqv_0 gamma gamma' e e' np e_eq e'_eq wf_np =>
          NP_eqv_0 _ _ _ gamma gamma' e e' np e_eq e'_eq wf_np
        | NP_eqv_1 gamma gamma' e e' np a a' e_eq e'_eq a_eqv wf_np =>
          NP_eqv_1 _ _ _ gamma gamma' e e' np a a' e_eq e'_eq (f _ a_eqv) wf_np
        | NP_eqv_2 gamma gamma' e e' np a a' b b' e_eq e'_eq a_eqv b_eqv wf_np =>
          NP_eqv_2 _ _ _ gamma gamma' e e' np a a' b b' e_eq e'_eq (f _ a_eqv) (f _ b_eqv) wf_np
        | NP_eqv_3 gamma gamma' e e' np a a' b b' c c' e_eq e'_eq a_eqv b_eqv c_eqv wf_np =>
          NP_eqv_3 _ _ _ gamma gamma' e e' np a a' b b' c c' e_eq e'_eq (f _ a_eqv) (f _ b_eqv) (f _ c_eqv) wf_np
      end.

    Global Instance iFun_Arith_eqv A B : iFunctor (NP_Functor_eqv A B).
      constructor 1 with (ifmap := NP_Functor_eqv_ifmap A B).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; reflexivity.
    Defined.

    (* Projection doesn't affect Equivalence Relation.*)

     Definition EQV_proj1_P A B (i : eqv_i A B) :=
      forall a' b' H H0, a' = proj1_sig (eqv_a _ _ i) ->
        b' = proj1_sig (eqv_b _ _ i) ->
        E_eqvC (env_A _ _ i) (env_B _ _ i) (exist _ a' H) (exist _ b' H0).

     Inductive EQV_proj1_Name := eqv_proj1_name.
     Context {EQV_proj1_EQV : forall A B,
       iPAlgebra EQV_proj1_Name (@EQV_proj1_P A B) (EQV_E A B)}.
     Context {Fun_EQV_E : forall A B, iFunctor (EQV_E A B)}.

     Definition EQV_proj1 A B:=
       ifold_ (EQV_E A B) _ (ip_algebra (iPAlgebra := EQV_proj1_EQV A B)).

    Variable Sub_NP_Functor_eqv_EQV_E : forall A B,
      Sub_iFunctor (NP_Functor_eqv A B) (EQV_E A B).

     Global Instance EQV_proj1_NP_Functor_eqv :
       forall A B,
         iPAlgebra EQV_proj1_Name (EQV_proj1_P A B) (NP_Functor_eqv _ _).
     Proof.
       intros; econstructor; unfold iAlgebra; intros.
       eapply ind_alg_NP_Functor_eqv; unfold EQV_proj1_P; simpl; intros.
       apply inject_i; econstructor; simpl; eauto.
       rewrite H2; rewrite e_eq; eauto.
       rewrite H3; rewrite e'_eq; eauto.
       apply inject_i; econstructor 2; simpl; eauto.
       rewrite H2; rewrite e_eq; eauto.
       rewrite H3; rewrite e'_eq; eauto.
       destruct a; destruct a'; apply IHa; auto.
       apply inject_i; econstructor 3; simpl; eauto.
       rewrite H2; rewrite e_eq; eauto.
       rewrite H3; rewrite e'_eq; eauto.
       destruct a; destruct a'; apply IHa; auto.
       destruct b; destruct b'; apply IHb; auto.
       apply inject_i; econstructor 4; simpl; eauto.
       rewrite H2; rewrite e_eq; eauto.
       rewrite H3; rewrite e'_eq; eauto.
       destruct a; destruct a'; apply IHa; auto.
       destruct b; destruct b'; apply IHb; auto.
       destruct c; destruct c'; apply IHc; auto.
       assumption.
     Defined.

   End eqv_Section.

   Variable EQV_E : forall A B, (@eqv_i A B -> Prop) -> eqv_i A B -> Prop.
   Context {Fun_EQV_E : forall A B, iFunctor (EQV_E A B)}.
   Variable WFV : (WFValue_i D V TypContext -> Prop) -> WFValue_i D V TypContext -> Prop.
   Context {funWFV : iFunctor WFV}.

    Class GammaTypContextC (TypContext : Set) : Set :=
      { Gamma : TypContext -> list (DType D);
        GammaSet : TypContext -> list (DType D) -> TypContext;
        GammaInsert : DType D -> TypContext -> TypContext;
        Gamma_GammaSet : forall Sigma env, 
          Gamma (GammaSet Sigma env) = env;
        GammaSet_Gamma : forall Sigma, 
          GammaSet Sigma (Gamma Sigma) = Sigma;
        Gamma_Insert : forall Sigma gamma T,
          Gamma Sigma = gamma -> 
          Gamma (GammaInsert T Sigma) = insert _ T gamma
      }.

    Class GammaConsExtensionC (TypContext : Set) 
      (ConsExt : ConsExtensionC TypContext)
      (GammaTyp : GammaTypContextC TypContext) : Type :=
      { 
        ConsExtension_Gamma : forall Sigma Sigma',
          ConsExtension Sigma' Sigma -> 
          Gamma Sigma = Gamma Sigma'}.

    Context {GammaTypContext : GammaTypContextC TypContext}.
    Context {TypContext_GCE : GammaConsExtensionC TypContext _ _}.

   Definition WF_eqv_environment_P
     (gamma_gamma' : prod (Env nat) (Env (DType D))) (Sigma : TypContext) :=
     (forall m b : nat,
       lookup (fst gamma_gamma') m = Some b ->
       exists T, lookup (snd gamma_gamma') b = Some T) /\
     Datatypes.length (snd gamma_gamma') = Datatypes.length (fst gamma_gamma') /\
     (forall m b : nat, lookup (fst gamma_gamma') m = Some b -> b = m) /\
     (Gamma Sigma = snd gamma_gamma').

   Lemma WF_eqv_environment_P_Sub : forall gamma_gamma' Sigma Sigma',
        WF_eqv_environment_P gamma_gamma' Sigma -> 
        ConsExtension Sigma' Sigma -> 
        WF_eqv_environment_P gamma_gamma' Sigma'.
   Proof.
     intros; destruct H as [WF_gamma [WF_gamma2 [WF_gamma' WF_gamma'']]].
     repeat split; eauto.
     intros; apply sym_eq; erewrite <- WF_gamma'', ConsExtension_Gamma; eauto.
   Qed.

  (*
    Definition eqv_eval_alg_Soundness'_P
      (typeof_rec : Exp (E (typeofR D MT)) -> typeofR D MT)
      (eval_rec : Exp (E nat) -> evalR V)
      (typeof_F : Mixin (Exp (E (typeofR D MT))) (E (typeofR D MT)) (typeofR D MT))
      (eval_F : Mixin (Exp (E nat)) (E nat) (evalR V))
      i :=
      E_eqv EQV_E _ _ i /\
      eval_alg_Soundness_P D V (E nat) WFV
      _ WF_eqv_environment_P
      (E (typeofR D)) _ (env_A _ _ i, env_B _ _ i) typeof_rec eval_rec
      typeof_F eval_F
      (proj1_sig (eqv_a _ _ i), proj1_sig (eqv_b _ _ i))
      (conj (proj2_sig (eqv_a _ _ i)) (proj2_sig (eqv_b _ _ i))).
    *)

    Lemma WF_eqv_environment_P_insert : forall Sigma gamma gamma' T,
      WF_eqv_environment_P (gamma, gamma') Sigma ->
      WF_eqv_environment_P (insert _ (Datatypes.length gamma) gamma, insert _ T gamma') 
      (GammaInsert T Sigma).
    Proof.
      intros; destruct H as [WF_gamma [WF_gamma2 [WF_gamma' WF_gamma'']]].
      unfold WF_eqv_environment_P; simpl in *|-*; repeat split; auto.
      rewrite <- WF_gamma2.
      revert WF_gamma; clear; simpl; induction gamma;
        destruct m; simpl; intros; try discriminate.
      injection H; intros; subst.
      clear; induction gamma'; simpl; eauto; eexists.
      injection H; intros; subst.
      generalize b (WF_gamma 0 _ (eq_refl _)); clear; induction gamma'; simpl; intros b H;
        destruct H as [T' lookup_T']; try discriminate.
      destruct b; eauto.
      eapply IHgamma.
      intros n0 b0 H0; eapply (WF_gamma (S n0) _ H0).
      eassumption.
      assert (exists m', Datatypes.length gamma = m') as m'_eq
        by (eexists _; reflexivity); destruct m'_eq as [m' m'_eq].
      rewrite m'_eq; generalize m' gamma WF_gamma2; clear; induction gamma';
        destruct gamma; intros; simpl; try discriminate;
          try injection H7; intros; eauto.
      simpl in *|-*.
      intro; caseEq (beq_nat m (Datatypes.length gamma)).
      assert (exists m', m' = Datatypes.length gamma) as ex_m' by
        (eexists _; reflexivity); destruct ex_m' as [m' m'_eq];
          rewrite <- m'_eq in H at 1.
      rewrite <- WF_gamma2 in H0. 
      rewrite (beq_nat_true _ _ H).
      rewrite (beq_nat_true _ _ H), m'_eq in H0. 
      rewrite <- WF_gamma2 in m'_eq; rewrite m'_eq.
      generalize m' b H0; clear.
      induction gamma; simpl; intros; try discriminate.
      injection H0; auto.
      eauto.
      eapply WF_gamma'.
      rewrite <- WF_gamma2 in H0.
      assert (exists m', m' = Datatypes.length gamma) as ex_m' by
      (eexists _; reflexivity); destruct ex_m' as [m' m'_eq].
      generalize m' m (beq_nat_false _ _ H) H0; clear;
        induction gamma; simpl; destruct m; intros;
          try discriminate; eauto.
      elimtype False; eauto.
      intros; eapply Gamma_Insert; eauto.
    Qed. 

    Definition WF_eqv_environment'_P e e'
     (gamma_gamma' : prod (Env nat) (Env (DType D))) (Sigma : TypContext) :=
     E_eqvC EQV_E (fst gamma_gamma') (snd gamma_gamma') e e' /\
     WF_eqv_environment_P gamma_gamma' Sigma.

   Lemma WF_eqv_environment'_P_Sub : forall e e' gamma_gamma' Sigma Sigma',
        WF_eqv_environment'_P e e' gamma_gamma' Sigma -> 
        ConsExtension Sigma' Sigma ->
        WF_eqv_environment'_P e e' gamma_gamma' Sigma' .
   Proof.
     intros; destruct H as [WF_gamma [WF_gamma2 [WF_gamma' [WF_gamma'' WF_Sigma]]]].
     repeat split; eauto.
     intros; apply sym_eq; erewrite <- WF_Sigma, ConsExtension_Gamma; eauto.
   Qed.

    Variable WFVM : (WFValueM_i D V MT ME TypContext -> Prop) -> 
      WFValueM_i D V MT ME TypContext -> Prop.
    Variable funWFVM : iFunctor WFVM.

    Definition eqv_soundness_X_P i :=
      forall n
      (IH : forall gamma gamma' e'  e'',
        E_eqvC EQV_E gamma' gamma e'' e' ->
        forall Sigma 
          (WF_gamma : forall n b, lookup gamma' n = Some b ->
            exists T, lookup gamma b = Some T)
          (WF_gamma2 : List.length gamma = List.length gamma')
          (WF_gamma' : forall n b, lookup gamma' n = Some b -> b = n)
          (WF_Sigma : Gamma Sigma = gamma),
          WFValueMC _ _ MT ME _ WFVM Sigma (bevalM n e'')
          (typeof D (E (DType D)) MT (Typeof_E := Typeof_E) (proj1_sig e'))),
      E_eqv EQV_E _ _ i /\
      forall Sigma 
        (WF_gamma : forall n b, lookup (env_A _ _ i) n = Some b ->
          exists T, lookup (env_B _ _ i) b = Some T)
        (WF_gamma2 : List.length (env_B _ _ i) = List.length (env_A _ _ i))
        (WF_gamma' : forall n b, lookup (env_A _ _ i) n = Some b -> b = n)
        (WF_Sigma : Gamma Sigma = env_B _ _ i),
        WFValueMC _ _ MT ME _ WFVM Sigma (bevalM (S n) (eqv_a _ _ i))
        (typeof D (E (DType D)) MT (Typeof_E := Typeof_E) (proj1_sig (eqv_b _ _ i))).

    Inductive eqv_soundness_X_Name : Set := eqv_soundness_X_name.

    Context {eqv_soundness_X_alg :
      iPAlgebra eqv_soundness_X_Name eqv_soundness_X_P (EQV_E _ _)}.
    Context {Sub_WFVM_Bot_WFVM : Sub_iFunctor WFValueM_Bot WFVM}.

    Lemma eval_soundness_X :
      forall n gamma gamma' e'  e'',
        E_eqvC EQV_E gamma' gamma e'' e' ->
        forall Sigma
          (WF_gamma : forall n b, lookup gamma' n = Some b ->
            exists T, lookup gamma b = Some T)
          (WF_gamma2 : List.length gamma = List.length gamma')
          (WF_gamma' : forall n b, lookup gamma' n = Some b -> b = n)
          (WF_Sigma : Gamma Sigma = gamma),
          WFValueMC _ _ MT ME _ WFVM Sigma (bevalM n e'')
          (typeof D (E (DType D)) MT (Typeof_E := Typeof_E) (proj1_sig e')).
    Proof.
      induction n; simpl.
      (* Base case. *)
      unfold bevalM; simpl; intros; apply (inject_i (subGF := Sub_WFVM_Bot_WFVM)); constructor.
      (* Inductive case. *)
      intros gamma gamma' e' e'' e'_e''_eqv; revert n IHn.
      eapply (ifold_ (EQV_E _ _) _ (ip_algebra (iPAlgebra := eqv_soundness_X_alg)) _ e'_e''_eqv).
    Qed.

    Definition eqv_soundness_X'_P 
      (typeof_rec : Exp (E (DType D)) -> typeofR D MT)
      (eval_rec : Exp (E nat) -> evalMR V ME)
      (typeof_F : Mixin (Exp (E (DType D))) (E (DType D)) (typeofR D MT))
      (eval_F : Mixin (Exp (E nat)) (E nat) (evalMR V ME))
      (i : eqv_i nat (DType D)) :=
      E_eqv EQV_E _ _ i /\
      eval_soundness'_P D V (E nat) MT ME _ WFVM
      _ (E (DType D)) _ WF_eqv_environment'_P 
      (env_A _ _ i, env_B _ _ i) typeof_rec eval_rec 
      typeof_F eval_F
      (proj1_sig (eqv_b _ _ i), proj1_sig (eqv_a _ _ i))
      (conj (proj2_sig (eqv_b _ _ i)) (proj2_sig (eqv_a _ _ i))).

    Definition soundness_X'_P 
      (typeof_rec : Exp (E (DType D)) -> typeofR D MT)
      (eval_rec : Exp (E nat) -> evalMR V ME)
      (typeof_F : Mixin (Exp (E (DType D))) (E (DType D)) (typeofR D MT))
      (eval_F : Mixin (Exp (E nat)) (E nat) (evalMR V ME))
      (i : eqv_i nat (DType D)) :=
      forall (IH : forall (e : Exp _) (e' : Exp _)
        pb Sigma (WF_gamma'' : WF_eqv_environment'_P e e' pb Sigma),
        E_eqvC EQV_E (fst pb) (snd pb) e e' ->
        WFValueMC _ _ MT ME _ WFVM Sigma (eval_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig e))))
        (typeof_rec e')),
      E_eqv EQV_E _ _ i /\
      eval_soundness'_P D V (E nat) MT ME _ WFVM
      _ (E (DType D)) _ WF_eqv_environment'_P
      (env_A _ _ i, env_B _ _ i) typeof_rec eval_rec 
      typeof_F eval_F
      (proj1_sig (eqv_b _ _ i), proj1_sig (eqv_a _ _ i))
      (conj (proj2_sig (eqv_b _ _ i)) (proj2_sig (eqv_a _ _ i))).
       
    (* forall 
        (eval_rec_proj : forall e, eval_rec e = eval_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig e))))
        (typeof_rec_proj : forall e, typeof_rec e = typeof_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig e))))
        Sigma 
        (IH : forall (e : Exp _) (e' : Exp _),          
          (forall (WF_gamma : forall n b, lookup (env_A _ _ i) n = Some b ->
            exists T, lookup (env_B _ _ i) b = Some T)
          (WF_gamma2 : List.length (env_A _ _ i) = List.length (env_B _ _ i))
          (WF_gamma' : forall n b, lookup (env_A _ _ i) n = Some b -> b = n),
          WFValueMC D V MT ME WFVM Sigma (env_B _ _ i) (eval_F eval_rec (out_t_UP' _ _ (proj1_sig e))) 
          (typeof_F typeof_rec (out_t_UP' _ _ (proj1_sig e'))) ->
          WFValueMC D V MT ME WFVM Sigma (env_B _ _ i) (eval_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig e))))
          (typeof_rec e'))),
      E_eqv EQV_E _ _ i /\
      forall (WF_gamma : forall n b, lookup (env_A _ _ i) n = Some b ->
        exists T, lookup (env_B _ _ i) b = Some T)
      (WF_gamma2 : List.length (env_A _ _ i) = List.length (env_B _ _ i))
      (WF_gamma' : forall n b, lookup (env_A _ _ i) n = Some b -> b = n),
      WFValueMC D V MT ME WFVM Sigma (env_B _ _ i)  (eval_F eval_rec (out_t_UP' _ _ (proj1_sig (eqv_a _ _ i)))) 
      (typeof_F typeof_rec (out_t_UP' _ _ (proj1_sig (eqv_b _ _ i)))). *)

    Inductive eqv_soundness_X'_Name : Set := eqv_soundness_X'_name.
    Inductive soundness_X'_Name : Set := soundness_X'_name.

    Context {eqv_soundness_X'_alg :
      forall eval_rec, 
      iPAlgebra soundness_X'_Name 
      (soundness_X'_P (fun e => typeof _ _ MT (proj1_sig e)) eval_rec
        (f_algebra (FAlgebra := Typeof_E _))
        (f_algebra (FAlgebra := bevalM_E))) (EQV_E _ _)}.

    Global Instance Lift_soundness_X_alg 
      typeof_rec eval_rec typeof_alg eval_alg 
      EQV_G {fun_EQV_G : iFunctor EQV_G}
      {EQV_G_EQV_Alg : iPAlgebra eqv_soundness_X'_Name (eqv_soundness_X'_P typeof_rec 
        eval_rec typeof_alg eval_alg) EQV_G} :
        iPAlgebra soundness_X'_Name (soundness_X'_P 
          typeof_rec eval_rec typeof_alg eval_alg) EQV_G.
      intros; econstructor; generalize (ip_algebra); unfold iAlgebra; intros.
      unfold soundness_X'_P; intros.
      assert (EQV_G (eqv_soundness_X'_P typeof_rec eval_rec typeof_alg eval_alg) i).
      eapply ifmap; try eapply H0.
      intros; apply H1; apply IH.
      apply (H _ H1).
    Defined.

    Lemma soundness_X' :
      forall eval_rec gamma gamma' e' e'',
        E_eqvC EQV_E gamma gamma' e' e'' ->
        soundness_X'_P (fun e => typeof _ _ MT (proj1_sig e)) eval_rec
        (f_algebra (FAlgebra := Typeof_E _))
        (f_algebra (FAlgebra := bevalM_E)) (mk_eqv_i nat (DType D) gamma gamma' e' e'').
    Proof.
      intros; apply (ifold_ (EQV_E _ _ )); try assumption.
      apply ip_algebra.
    Qed.

    Context {WF_MAlg_typeof : WF_MAlgebra Typeof_E}.

    (* ============================================== *)
    (* Evaluation preserves Well-Formedness           *)
    (* ============================================== *)

    Variable SVM : (SubValueM_i V ME -> Prop) -> SubValueM_i V ME -> Prop.
    Context {funSVM : iFunctor SVM}.
    Context {Sub_SVM_base_SVM : Sub_iFunctor (SubValueM_base V ME) SVM}.

    Context {bevalM_continuous_Exp_E : 
      DAlgebra EC_ExpName (E nat) (E nat) (UP'_SP (bevalM_continuous_Exp_P V (E nat) ME SVM))}.

    (** Bounded Monadic Evaluation is continuous. **)

    Definition WFValueM_continuous_P (e : Fix (E nat))
      (e_UP' : Universal_Property'_fold e) :=
      forall (m : nat),
        (forall Sigma T (e0 e' : (Names.Exp (E nat))) (n : nat), m <= n ->
          proj1_sig e' = proj1_sig e0 ->
          WFValueMC _ _ MT ME _ WFVM Sigma (bevalM n e') T ->
          WFValueMC _ _ MT ME _ WFVM Sigma (bevalM m e') T) ->
        forall Sigma T (e' : Names.Exp (E nat)) (n : nat),
          m <= n ->
          proj1_sig e' = e ->
          WFValueMC _ _ MT ME _ WFVM Sigma (bevalM (S n) e') T -> 
          WFValueMC _ _ MT ME _ WFVM Sigma (bevalM (S m) e') T.

    Inductive WFVM_ContinuousName : Set := wfvm_continuousname.

    Context {WFValue_continuous_E : 
      DAlgebra WFVM_ContinuousName (E nat) (E nat) (UP'_P WFValueM_continuous_P)}.

    Lemma WF_Value_beval : forall m n,
      forall (e : Exp _) Sigma T,
        m <= n -> 
        WFValueMC _ _ MT ME _ WFVM Sigma (bevalM n e) T -> 
        WFValueMC _ _ MT ME _ WFVM Sigma (bevalM m e) T.
    Proof.
      intro; induction m; intros.
      unfold bevalM; apply inject_i; constructor.
      inversion H; subst; auto.
      destruct e as [e e_UP'].
      generalize (projT2 (IndD (P := UP'_SP WFValueM_continuous_P) e e_UP')).
      unfold WFValueM_continuous_P; intros; eapply H2 with (n := m0); eauto with arith.
    Qed.

    Context {EQV_proj1_EQV : forall A B,
      iPAlgebra EQV_proj1_Name (@EQV_proj1_P EQV_E A B) (EQV_E A B)}.

    Lemma eval_soundness_X' : forall n gamma gamma' e'  e'',
      E_eqvC EQV_E gamma gamma' e' e'' ->
      forall Sigma
        (WF_gamma : forall n b, lookup gamma n = Some b ->
          exists T, lookup gamma' b = Some T)
        (WF_gamma2 : List.length gamma' = List.length gamma)
        (WF_gamma' : forall n b, lookup gamma n = Some b -> b = n)
        (WF_Sigma : Gamma Sigma = gamma'),
        WFValueMC _ _ MT ME _ WFVM Sigma (bevalM n e')
        (typeof D (E (DType D)) MT (Typeof_E := Typeof_E) (proj1_sig e'')).
    Proof.
      induction n; simpl;
        intros; unfold beval; simpl in *|-*.
      apply (inject_i (subGF := Sub_WFVM_Bot_WFVM)); econstructor; eauto.
      generalize (soundness_X' (bevalM n) _ _ _ _ H).
      unfold soundness_X'_P; simpl; intros.
      unfold bevalM; intros; simpl. 
      rewrite <- (@in_out_UP'_inverse _ _ (proj1_sig e'') (proj2_sig _)).
      unfold typeof, evalM, mfold; simpl; unfold in_t; 
        repeat rewrite wf_malgebra; unfold mfold; apply H0; auto.      
      intros; unfold bevalM; erewrite bF_UP_in_out.
      instantiate (1 := proj2_sig _).
      destruct e; simpl; auto.
      unfold bevalM in IHn.
      destruct WF_gamma'' as [e'0_eqv [WF_gamma'' [WF_gamma''2 [WF_gamma''' Sub_Gamma_gamma'']]]].
      subst; eapply IHn; eauto.
      intros; unfold bevalM; erewrite bF_UP_in_out.
      instantiate (1 := proj2_sig _); destruct e; auto.
      intro; rewrite in_out_UP'_inverse; auto; exact (proj2_sig _).
      repeat split; auto.
      simpl; destruct e'; destruct e''; auto.
      simpl; eapply (EQV_proj1 _ _ _ _ H); simpl; auto.
      intros;
        destruct WF_Gamma as [e'0_eqv [WF_gamma'' [WF_gamma''2 [WF_gamma''' Sub_Gamma_gamma'']]]];
          subst.
      eapply IHn with (gamma := fst pb) (gamma' := snd pb); eauto.
      simpl; destruct (fst a); destruct (snd a);
        apply (EQV_proj1 _ _ _ _ e'0_eqv); simpl; auto.
      generalize in_out_UP'_inverse; simpl; intros in_eq';
        apply in_eq'; auto.
    Qed.

    Definition WF_EnvG (env : list (Value V)) (Sigma : TypContext) : Prop :=
      P2_Env (WFValueC _ _ _ WFV Sigma) env (Gamma Sigma).

    Inductive wfvm_bind_Name : Set := wfvm_bind_name.

    Definition wfvm_bind_P (i : WFValueM_i D V MT ME _) :=
      forall ke kt
        (kt_proj1_eq : forall T T' : Names.DType D,
          proj1_sig T = proj1_sig T' -> fmap (@proj1_sig _ _) (kt T) = fmap (@proj1_sig _ _) (kt T'))
        (WFVM_ke_kt : forall Sigma' v T,
          ConsExtension Sigma' (wfvm_S _ _ _ _ _ i) -> 
          WFValueC _ _ _ WFV Sigma' v T ->
          WFValueMC _ _ MT ME _ WFVM Sigma'  (ke v) (kt T)),
        WFValueMC _ _ MT ME _ WFVM (wfvm_S _ _ _ _ _ i) (wfvm_v _ _ _ _ _ i >>= ke) (wfvm_T _ _ _ _ _ i >>= kt).

    Context {wfvm_bind_alg : iPAlgebra wfvm_bind_Name wfvm_bind_P WFVM}.
    Context {WFV_proj1_b_WFV : iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V _ WFV) WFV}.

    Lemma wfvm_bind Sigma (mv : ME (Value V)) (mT : MT (DType D))
          (ke : Value V -> ME (Value V)) (kt : DType D -> MT (DType D))
          (kt_proj1_eq : forall T T' : Names.DType D,
            proj1_sig T = proj1_sig T' -> fmap (@proj1_sig _ _) (kt T) = fmap (@proj1_sig _ _) (kt T'))
          (WFVM_mv_mT : WFValueMC _ _ MT ME _ WFVM Sigma mv mT)
          (WFVM_ke_kt : forall Sigma' v T,
            ConsExtension Sigma' Sigma -> 
            WFValueC _ _ _ WFV Sigma' v T -> WFValueMC _ _ MT ME _ WFVM Sigma' (ke v) (kt T)) :
       WFValueMC _ _ MT ME _ WFVM Sigma (mv >>= ke) (mT >>= kt).
    Proof.
      eapply (ifold_ WFVM _ (ip_algebra (iPAlgebra := wfvm_bind_alg)) (mk_WFValueM_i _ _ MT ME _ Sigma mv mT) WFVM_mv_mT);
        eauto.
    Qed.

    Lemma wfvm_bind' Sigma (mv : ME (Value V)) (mT : MT (DType D))
          (ke : Value V -> ME (Value V)) (kt : DType D -> MT (DType D))
          (kt_proj1_eq : forall T T' : Names.DType D,
            proj1_sig T = proj1_sig T' -> fmap (@proj1_sig _ _) (kt T) = fmap (@proj1_sig _ _) (kt T'))
          (WFVM_mv_mT : WFValueMC _ _ MT ME _ WFVM Sigma mv mT)
          (WFVM_ke_kt : forall Sigma' v T,
            ConsExtension Sigma' Sigma -> 
            WFValueC _ _ _ WFV Sigma' v T -> WFValueMC _ _ MT ME _ WFVM Sigma' (ke v) (kt T)) :
       WFValueMC _ _ MT ME _ WFVM Sigma (mv >>= ke) (match mT with
                                                           | Some T => kt T
                                                           | None => None
                                                         end).
    Proof.
      generalize wfvm_bind; simpl; auto.
    Qed.

    Context {Sub_WFVM_Base_WFVM : Sub_iFunctor (WFValueM_base D V MT ME _ WFV) WFVM}.

    Global Instance wfvm_bind_base' :
      iPAlgebra wfvm_bind_Name wfvm_bind_P (WFValueM_base D V MT ME _ WFV).
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base with (WFV := WFV)
        (Fail_MT := _) (Monad_ME := _);
        try assumption; unfold wfvm_bind_P; simpl; intros.
      (* WFVM_Return' *)
      destruct mt.
      rewrite <- left_unit; eapply WFVM_ke_kt; eauto.
      apply ConsExtension_id.
      destruct H1 as [mt' mt'_eq]; subst.
      destruct mt'; try discriminate; injection mt'_eq; intros; subst.
      destruct s; simpl.
      destruct x; apply (WFV_proj1_b D V _ WFV _ _ H0); simpl; auto.
      generalize (WFVM_Untyped D V MT ME _ WFV (iFix WFVM) _ Sigma
        (return_ v >>= ke) (fail (A := DType D))).
      apply inject_i; simpl; auto.
      (* WFVM_Untyped' *)
      generalize (WFVM_Untyped D V MT ME _ WFV (iFix WFVM) _ Sigma
        (me >>= ke) (fail (A := DType D))).
      destruct mt; simpl; apply inject_i; auto.
    Qed.

    Global Instance wfvm_bind_Bot :
      iPAlgebra wfvm_bind_Name wfvm_bind_P WFValueM_Bot.
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_Bot;
        try assumption; unfold wfvm_bind_P; simpl; intros.
    (* WFVM_fail *)
      rewrite bind_fail; apply (inject_i (subGF := Sub_WFVM_Bot_WFVM)); constructor.
    Qed.

    Section NP_beval_Soundness.

    Variable (NP : Set -> Set).
    Context {Fun_NP : Functor NP}.
    Context {sub_NP_F : forall A, NP :<: E A}.
    Context {WF_sub_NP_F : forall A, WF_Functor _ _ (sub_NP_F A) _ _}.

    Context {Sub_NP_Functor_eqv_EQV_E : forall A B,
      Sub_iFunctor (NP_Functor_eqv NP A B) (EQV_E A B)}.

    Definition EQV_NP_invert_P A B (i : eqv_i A B) :=
      E_eqv EQV_E _ _ i /\
      forall b : NP _, proj1_sig (eqv_b _ _ i) = inject b ->
        NP_Functor_eqv NP A B (E_eqv EQV_E _ _) i.

     Inductive EQV_NP_invert_Name := eqv_NP_invert_name.

     Global Instance EQV_NP_invert_EQV_NP : forall A B,
       iPAlgebra EQV_NP_invert_Name (@EQV_NP_invert_P A B) (NP_Functor_eqv NP A B).
     Proof.
       intros; econstructor; unfold iAlgebra; intros.         
       eapply ind_alg_NP_Functor_eqv; try eassumption; 
         unfold EQV_NP_invert_P; simpl; intros; split.
       eapply inject_i; econstructor; eauto.
       intros; econstructor; eauto.
       destruct IHa as [IHa IHa']; eapply inject_i; econstructor 2; eauto.
       destruct IHa as [IHa IHa']; intros; econstructor 2; eauto.
       destruct IHa as [IHa IHa']; destruct IHb as [IHb IHb']; 
         eapply inject_i; econstructor 3; eauto.
       destruct IHa as [IHa IHa']; destruct IHb as [IHb IHb']; 
         intros; econstructor 3; eauto.
       destruct IHa as [IHa IHa']; destruct IHb as [IHb IHb']; 
         destruct IHc as [IHc IHc']; eapply inject_i; econstructor 4; eauto.
       destruct IHa as [IHa IHa']; destruct IHb as [IHb IHb']; 
         destruct IHc as [IHc IHc']; intros; econstructor 4; eauto.
     Defined.

     Context {Except_invert_EQV_E : forall A B,
       iPAlgebra EQV_NP_invert_Name (@EQV_NP_invert_P A B) (EQV_E A B)}.

     Definition EQV_NP_invert A B :=
       ifold_ (EQV_E A B) _ (ip_algebra (iPAlgebra := Except_invert_EQV_E _ _)).

     Global Instance EQV_NP_invert_EQV_NP' {NP' : Set -> Set}
       {Fun_NP' : Functor NP'}
       {sub_NP'_F : forall A, NP' :<: E A}
       {WF_sub_NP'_F : forall A, WF_Functor _ _ (sub_NP'_F A) _ _}
       {Sub_NP'_Functor_eqv_EQV_E : forall A B,
         Sub_iFunctor (NP_Functor_eqv NP' A B) (EQV_E A B)}
       {Dis_NP_NP' : forall A, Distinct_Sub_Functor (Fun_E A) (sub_NP_F A) (sub_NP'_F A)}
       : forall A B,
       iPAlgebra EQV_NP_invert_Name (@EQV_NP_invert_P A B) (NP_Functor_eqv NP' A B).
     Proof.
       intros; econstructor; unfold iAlgebra; intros.         
       apply ind_alg_NP_Functor_eqv with (NP := NP')
         (Fun_NP := Fun_NP') (sub_NP_F := sub_NP'_F); try eassumption;
           unfold EQV_NP_invert_P; simpl; intros; split.
       eapply (inject_i (subGF := Sub_NP'_Functor_eqv_EQV_E _ _)); econstructor; eauto.
       intros; rewrite e'_eq in H0.
       elimtype False; eapply (inject_discriminate (WF_F := WF_sub_NP_F B)
       (WF_G := WF_sub_NP'_F B) (Dis_NP_NP' _)); eauto.
       destruct IHa as [IHa IHa']; eapply inject_i; econstructor 2; eauto.
       intros; rewrite e'_eq in H0.
       elimtype False; eapply (inject_discriminate (WF_F := WF_sub_NP_F B)
       (WF_G := WF_sub_NP'_F B) (Dis_NP_NP' _)); eauto.
       destruct IHa as [IHa IHa']; destruct IHb as [IHb IHb']; 
         eapply inject_i; econstructor 3; eauto.
       intros; rewrite e'_eq in H0.
       elimtype False; eapply (inject_discriminate (WF_F := WF_sub_NP_F B)
       (WF_G := WF_sub_NP'_F B) (Dis_NP_NP' _)); eauto.
       destruct IHa as [IHa IHa']; destruct IHb as [IHb IHb']; 
         destruct IHc as [IHc IHc']; eapply inject_i; econstructor 4; eauto.
       intros; rewrite e'_eq in H0.
       elimtype False; eapply (inject_discriminate (WF_F := WF_sub_NP_F B)
       (WF_G := WF_sub_NP'_F B) (Dis_NP_NP' _)); eauto.
     Defined.

    Context {eval_soundness'_Exp_NP : forall
      (P_Sub : forall e e' gamma_gamma' Sigma Sigma', 
        WF_eqv_environment'_P e e' gamma_gamma' Sigma -> 
        ConsExtension Sigma' Sigma -> 
        WF_eqv_environment'_P e e' gamma_gamma' Sigma' )
      gamma_gamma' typeof_rec eval_rec, 
      P2Algebra ES'_ExpName NP (E _) (E _)
      (UP'_P2 (eval_soundness'_P D V _ MT ME _ WFVM 
        _ (E (DType D)) _ WF_eqv_environment'_P gamma_gamma' typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_E _)) (f_algebra (FAlgebra := bevalM_E))))}.

    Context {Typeof_NP : forall T, FAlgebra TypeofName T (typeofR D MT) NP}.
    Context {beval_NP : FAlgebra EvalName (Exp (E nat)) (evalMR V ME) NP}.
    Context {WF_eval_F : @WF_FAlgebra EvalName _ _ NP (E _)
      (sub_NP_F _) beval_NP (bevalM_E)}.
    Context {WF_typeof_F : @WF_FAlgebra TypeofName _ _ NP (E _)
      (sub_NP_F _) (Typeof_NP _) (Typeof_E (Fix (E (typeofR D MT))))}.

    Global Instance eqv_eval_Soundness typeof_rec eval_rec :
      iPAlgebra eqv_soundness_X'_Name 
      (eqv_soundness_X'_P typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_E _))
        (f_algebra (FAlgebra := bevalM_E))) (NP_Functor_eqv _ _ _).
    Proof.
      econstructor; unfold iAlgebra; intros.
      eapply ind_alg_NP_Functor_eqv; try eassumption;
        unfold eqv_soundness_X'_P; simpl; intros.
      split.
      apply inject_i; econstructor; eauto.
      generalize (proj1_WF_Ind (P2Algebra := eval_soundness'_Exp_NP 
        WF_eqv_environment'_P_Sub (gamma, gamma') typeof_rec eval_rec) (np _)); simpl.
      generalize (proj2_WF_Ind (P2Algebra := eval_soundness'_Exp_NP
        WF_eqv_environment'_P_Sub (gamma, gamma') typeof_rec eval_rec) (np _)); simpl.
      intros e1_eq e2_eq.
      destruct (p2_algebra (P2Algebra := eval_soundness'_Exp_NP 
        WF_eqv_environment'_P_Sub (gamma, gamma') typeof_rec eval_rec)) as
        [[e1 e2] [[UP_e1 UP_e2] sound_e1]]; auto; simpl in *|-*.
      destruct e as [e e_UP]; destruct e' as [e' e'_UP]; simpl.
      simpl in *|-*.
      revert sound_e1.
      unfold eval_soundness'_P; simpl.
      repeat rewrite e_eq, e'_eq, e1_eq, e2_eq; simpl.
      intros sound_e1 typeof_rec_proj eval_rec_proj Sigma WF_Gamma IHa0.
      assert (WF_eqv_environment'_P
        (exist Universal_Property'_fold e2 (proj2 (conj UP_e1 UP_e2)))
        (exist Universal_Property'_fold e1 (proj1 (conj UP_e1 UP_e2)))
        (gamma, gamma') Sigma) as WF'_Gamma.
      split; destruct WF_Gamma; simpl; eauto.
      apply inject_i; constructor 1 with (np := np).
      simpl; rewrite e1_eq; unfold inject; simpl.
      rewrite wf_functor; apply f_equal; apply f_equal.
      apply wf_np.
      simpl; rewrite e2_eq; unfold inject; simpl.
      rewrite wf_functor; apply f_equal; apply f_equal.
      apply wf_np.
      auto. 
      generalize (sound_e1 typeof_rec_proj eval_rec_proj Sigma WF'_Gamma IHa0); intros.
      repeat rewrite e_eq, e'_eq; simpl.
      unfold inject; simpl.
      repeat rewrite out_in_fmap.
      repeat rewrite wf_functor; simpl.
      rewrite e1_eq, e2_eq in H0.
      repeat rewrite out_in_fmap in H0.
      repeat rewrite wf_functor in H0; simpl in H0.
      erewrite wf_np.
      erewrite wf_np with (A' := (@sig (Fix (E (@DType D Fun_D))) _))
      (f := @proj1_sig _ _).
      apply H0.
      destruct IHa as [eqv_a IHa]; split; intros.
      apply inject_i; econstructor 2; eauto.
      assert (UP'_P2
        (eval_soundness'_P D V _ MT ME _ WFVM 
        _ (E (DType D)) _ WF_eqv_environment'_P (gamma, gamma') typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_E _)) (f_algebra (FAlgebra := bevalM_E)))
        (proj1_sig a', proj1_sig a)).
      unfold UP'_P2; intros.
      econstructor.
      instantiate (1 := conj (proj2_sig _) (proj2_sig _)).
      unfold eval_soundness'_P; intros.
      simpl; apply IHa; auto.
      generalize (proj1_WF_Ind (P2Algebra := eval_soundness'_Exp_NP
        WF_eqv_environment'_P_Sub (gamma, gamma') typeof_rec eval_rec) 
      (np _ (exist _ (proj1_sig a', proj1_sig a) H0))); simpl.
      generalize (proj2_WF_Ind (P2Algebra := eval_soundness'_Exp_NP
        WF_eqv_environment'_P_Sub
        (gamma, gamma') typeof_rec eval_rec) 
      (np _ (exist _ (proj1_sig a', proj1_sig a) H0))); simpl.
      intros e1_eq e2_eq.
      destruct (p2_algebra (P2Algebra := eval_soundness'_Exp_NP 
        WF_eqv_environment'_P_Sub (gamma, gamma') typeof_rec eval_rec)) as
        [[e1 e2] [[UP_e1 UP_e2] sound_e1]]; auto; simpl in *|-*.
      destruct e as [e e_UP]; destruct e' as [e' e'_UP]; simpl.
      simpl in *|-*.
      revert sound_e1.
      unfold eval_soundness'_P; simpl.
      repeat rewrite e_eq, e'_eq, e1_eq, e2_eq; simpl.
      intros sound_e1 typeof_rec_proj eval_rec_proj Sigma WF_Gamma IHa0.
      assert (WF_eqv_environment'_P
        (exist Universal_Property'_fold e2 (proj2 (conj UP_e1 UP_e2)))
        (exist Universal_Property'_fold e1 (proj1 (conj UP_e1 UP_e2)))
        (gamma, gamma') Sigma) as WF'_Gamma.
      split; destruct WF_Gamma; simpl; eauto.
      apply inject_i; constructor 2 with (np := np) (a := a) (a' := a').
      simpl; rewrite e1_eq; unfold inject; simpl.
      rewrite wf_functor; apply f_equal; apply f_equal.
      apply wf_np; simpl; auto.
      simpl; rewrite e2_eq; unfold inject; simpl.
      rewrite wf_functor; apply f_equal; apply f_equal.
      apply wf_np; simpl; auto.
      auto.
      auto.
      generalize (sound_e1 typeof_rec_proj eval_rec_proj Sigma WF'_Gamma IHa0); intros.
      unfold inject; simpl.
      repeat rewrite e_eq, e'_eq; simpl.
      unfold inject; simpl.
      repeat rewrite out_in_fmap.
      repeat rewrite wf_functor; simpl.
      rewrite e1_eq, e2_eq in H1.
      repeat rewrite out_in_fmap in H1.
      repeat rewrite wf_functor in H1; simpl in H1.
      erewrite wf_np.
      erewrite wf_np with (A' := (@sig (Fix (E (@DType D Fun_D))) _))
      (f := @proj1_sig _ _).
      apply H1.
      simpl; auto.
      simpl; auto.
      destruct IHa as [eqv_a IHa]; destruct IHb as [eqv_b IHb]; split; intros.
      apply inject_i; econstructor 3; eauto.
      assert (UP'_P2
        (eval_soundness'_P D V _ MT ME _ WFVM 
        _ (E (DType D)) _ WF_eqv_environment'_P (gamma, gamma') typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_E _)) (f_algebra (FAlgebra := bevalM_E)))
        (proj1_sig a', proj1_sig a)).
      unfold UP'_P2; intros.
      econstructor.
      instantiate (1 := conj (proj2_sig _) (proj2_sig _)).
      unfold eval_soundness'_P; intros.
      simpl; apply IHa; auto.
      assert (UP'_P2
        (eval_soundness'_P D V _ MT ME _ WFVM 
        _  (E (DType D)) _ WF_eqv_environment'_P (gamma, gamma') typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_E _)) (f_algebra (FAlgebra := bevalM_E)))
        (proj1_sig b', proj1_sig b)).
      unfold UP'_P2; intros.
      econstructor.
      instantiate (1 := conj (proj2_sig _) (proj2_sig _)).
      unfold eval_soundness'_P; intros.
      simpl; apply IHb; auto.
      generalize (proj1_WF_Ind (P2Algebra := eval_soundness'_Exp_NP
        WF_eqv_environment'_P_Sub (gamma, gamma') typeof_rec eval_rec) 
      (np _ (exist _ (proj1_sig a', proj1_sig a) H0) (exist _ (proj1_sig b', proj1_sig b) H1))); simpl.
      generalize (proj2_WF_Ind (P2Algebra := eval_soundness'_Exp_NP
        WF_eqv_environment'_P_Sub (gamma, gamma') typeof_rec eval_rec)
      (np _ (exist _ (proj1_sig a', proj1_sig a) H0) (exist _ (proj1_sig b', proj1_sig b) H1))); simpl.
      intros e1_eq e2_eq.
      destruct (p2_algebra (P2Algebra := eval_soundness'_Exp_NP 
        WF_eqv_environment'_P_Sub (gamma, gamma') typeof_rec eval_rec)) as
        [[e1 e2] [[UP_e1 UP_e2] sound_e1]]; auto; simpl in *|-*.
      destruct e as [e e_UP]; destruct e' as [e' e'_UP]; simpl.
      simpl in *|-*.
      revert sound_e1.
      unfold eval_soundness'_P; simpl.
      simpl.
      intros sound_e1 typeof_rec_proj eval_rec_proj Sigma WF_Gamma IHa0.
      assert (WF_eqv_environment'_P
        (exist Universal_Property'_fold e2 (proj2 (conj UP_e1 UP_e2)))
        (exist Universal_Property'_fold e1 (proj1 (conj UP_e1 UP_e2)))
        (gamma, gamma') Sigma) as WF'_Gamma.
      split; destruct WF_Gamma; simpl; eauto.
      apply inject_i; constructor 3 with (np := np) (a := a) (a' := a') (b := b) (b' := b').
      simpl; rewrite e1_eq; unfold inject; simpl.
      rewrite wf_functor; apply f_equal; apply f_equal.
      apply wf_np; simpl; auto.
      simpl; rewrite e2_eq; unfold inject; simpl.
      rewrite wf_functor; apply f_equal; apply f_equal.
      apply wf_np; simpl; auto.
      auto.
      auto.
      auto.
      generalize (sound_e1 typeof_rec_proj eval_rec_proj Sigma WF'_Gamma IHa0); intros.
      repeat rewrite e_eq, e'_eq; simpl.
      unfold inject; simpl.
      repeat rewrite out_in_fmap.
      repeat rewrite wf_functor; simpl.
      rewrite e1_eq, e2_eq in H2.
      repeat rewrite out_in_fmap in H2.
      repeat rewrite wf_functor in H2; simpl in H2.
      erewrite wf_np.
      erewrite wf_np with (A' := (@sig (Fix (E (@DType D Fun_D))) _))
      (f := @proj1_sig _ _).
      apply H2.
      simpl; auto.
      simpl; auto.
      simpl; auto.
      simpl; auto.
      destruct IHa as [eqv_a IHa]; destruct IHb as [eqv_b IHb]; 
        destruct IHc as [eqv_c IHc]; split; intros.
      apply inject_i; econstructor 4; eauto.
      assert (UP'_P2
        (eval_soundness'_P D V _ MT ME _ WFVM 
        _ (E (DType D)) _ WF_eqv_environment'_P (gamma, gamma') typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_E _)) (f_algebra (FAlgebra := bevalM_E)))
        (proj1_sig a', proj1_sig a)).
      unfold UP'_P2; intros.
      econstructor.
      instantiate (1 := conj (proj2_sig _) (proj2_sig _)).
      unfold eval_soundness'_P; intros.
      simpl; apply IHa; auto.
      assert (UP'_P2
        (eval_soundness'_P D V _ MT ME _ WFVM 
        _ (E (DType D)) _ WF_eqv_environment'_P (gamma, gamma') typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_E _)) (f_algebra (FAlgebra := bevalM_E)))
        (proj1_sig b', proj1_sig b)).
      unfold UP'_P2; intros.
      econstructor.
      instantiate (1 := conj (proj2_sig _) (proj2_sig _)).
      unfold eval_soundness'_P; intros.
      simpl; apply IHb; auto.
      assert (UP'_P2
        (eval_soundness'_P D V _ MT ME _ WFVM 
        _ (E (DType D)) _ WF_eqv_environment'_P (gamma, gamma') typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_E _)) (f_algebra (FAlgebra := bevalM_E)))
        (proj1_sig c', proj1_sig c)).
      unfold UP'_P2; intros.
      econstructor.
      instantiate (1 := conj (proj2_sig _) (proj2_sig _)).
      unfold eval_soundness'_P; intros.
      simpl; apply IHc; auto.
      generalize (proj1_WF_Ind (P2Algebra := eval_soundness'_Exp_NP WF_eqv_environment'_P_Sub
        (gamma, gamma') typeof_rec eval_rec) (np _ (exist _ (proj1_sig a', proj1_sig a) H0) 
          (exist _ (proj1_sig b', proj1_sig b) H1) (exist _ (proj1_sig c', proj1_sig c) H2))); simpl.
      generalize (proj2_WF_Ind (P2Algebra := eval_soundness'_Exp_NP WF_eqv_environment'_P_Sub
        (gamma, gamma') typeof_rec eval_rec) (np _ (exist _ (proj1_sig a', proj1_sig a) H0)
          (exist _ (proj1_sig b', proj1_sig b) H1) (exist _ (proj1_sig c', proj1_sig c) H2))); simpl.
      intros e1_eq e2_eq.
      destruct (p2_algebra (P2Algebra := eval_soundness'_Exp_NP
        WF_eqv_environment'_P_Sub (gamma, gamma') typeof_rec eval_rec)) as
        [[e1 e2] [[UP_e1 UP_e2] sound_e1]]; auto; simpl in *|-*.
      destruct e as [e e_UP]; destruct e' as [e' e'_UP]; simpl.
      simpl in *|-*.
      revert sound_e1.
      unfold eval_soundness'_P; simpl.
      repeat rewrite e_eq, e'_eq, e1_eq, e2_eq; simpl.
      intros sound_e1 typeof_rec_proj eval_rec_proj Sigma WF_Gamma IHa0.
      assert (WF_eqv_environment'_P
        (exist Universal_Property'_fold e2 (proj2 (conj UP_e1 UP_e2)))
        (exist Universal_Property'_fold e1 (proj1 (conj UP_e1 UP_e2)))
        (gamma, gamma') Sigma) as WF'_Gamma.
      split; destruct WF_Gamma; simpl; eauto.
      apply inject_i; constructor 4 with 
        (np := np) (a := a) (a' := a') (b := b) (b' := b') (c := c) (c' := c').
      simpl; rewrite e1_eq; unfold inject; simpl.
      rewrite wf_functor; apply f_equal; apply f_equal.
      apply wf_np; simpl; auto.
      simpl; rewrite e2_eq; unfold inject; simpl.
      rewrite wf_functor; apply f_equal; apply f_equal.
      apply wf_np; simpl; auto.
      auto.
      auto.
      auto.
      auto.
      generalize (sound_e1 typeof_rec_proj eval_rec_proj Sigma WF'_Gamma IHa0); intros.
      rewrite e_eq, e'_eq.
      unfold inject; simpl.
      repeat rewrite out_in_fmap.
      repeat rewrite wf_functor; simpl.
      rewrite e1_eq, e2_eq in H3.
      repeat rewrite out_in_fmap in H3.
      repeat rewrite wf_functor in H3; simpl in H3.
      erewrite wf_np.
      erewrite wf_np with (A' := (@sig (Fix (E (@DType D Fun_D))) _))
      (f := @proj1_sig _ _).
      apply H3.
      simpl; auto.
      simpl; auto.
      simpl; auto.
      simpl; auto.
      simpl; auto.
      simpl; auto.
    Qed. 

    End NP_beval_Soundness.
  
  End WFValue_Sec.
  
  Section Pure_Sound_X_Sec.

    Context {Environment_ME : Environment ME (Env (Value V))}.

   Variable EQV_E : forall A B, (@eqv_i A B -> Prop) -> eqv_i A B -> Prop.
   Context {Fun_EQV_E : forall A B, iFunctor (EQV_E A B)}.
    Variable WFV : (WFValue_i D V (list (DType D)) -> Prop) -> WFValue_i D V (list (DType D)) -> Prop.
    Context {funWFV : iFunctor WFV}.
    Variable WFVM : (WFValueM_i D V MT ME (list (DType D)) -> Prop) -> 
      WFValueM_i D V MT ME (list (DType D)) -> Prop.
    Variable funWFVM : iFunctor WFVM.

    Global Instance GammaTypContext : GammaTypContextC (list (DType D)) :=
      {| Gamma := id;
        GammaSet := fun _ => id;
        GammaInsert := insert _|}.
    Proof.
      unfold id; intros; congruence.
      unfold id; intros; congruence.
      unfold id; intros; congruence.
    Defined.

    Global Instance GammaTypContextCE : ConsExtensionC (list (DType D)) :=
      {| ConsExtension := fun gamma gamma' => gamma = gamma' |}. 
    Proof.
      (* ConsExtension_id *) 
      eauto.
      (* ConsExtension_trans *) 
      intros; congruence.
    Defined.

    Global Instance GammaTypContextGCE : 
      GammaConsExtensionC (list (DType D)) (GammaTypContextCE)
      (GammaTypContext).
    Proof.
      econstructor; eauto.
    Defined.


    Definition Pure_Sound_X_P (i : WFValueM_i D V MT ME (list (DType D))) :=
      forall T gamma''
        (WF_gamma'' : WF_EnvG (GammaTypContext := GammaTypContext) _ WFV gamma'' (wfvm_S _ _ _ _ _ i)),
        wfvm_T _ _ _ _ _ i = return_ T ->
        local (fun _ => gamma'') (wfvm_v _ _ _ _ _ i) = fail \/
        exists v : Value V, 
          local (fun _ => gamma'') (wfvm_v _ _ _ _ _ i) = return_ (M := ME) v /\
          WFValueC _ _ _ WFV (wfvm_S _ _ _ _ _ i) v T.

    Inductive Pure_Sound_X_Name := pure_sound_X_name.
    
    Context {Pure_Sound_X_WFVM : iPAlgebra Pure_Sound_X_Name Pure_Sound_X_P WFVM}.
    Context {WFV_proj1_b_WFV : iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V _ WFV) WFV}.
    Context {EQV_proj1_EQV : forall A B,
      iPAlgebra EQV_proj1_Name (@EQV_proj1_P EQV_E A B) (EQV_E A B)}.
    Context {eqv_soundness_X'_alg :
      forall eval_rec, 
      iPAlgebra soundness_X'_Name 
      (soundness_X'_P _ EQV_E WFVM (fun e => typeof _ _ MT (proj1_sig e)) eval_rec
        (f_algebra (FAlgebra := Typeof_E _))
        (f_algebra (FAlgebra := bevalM_E))) (EQV_E _ _)}.
    Context {WF_MAlg_typeof : WF_MAlgebra Typeof_E}.
    Context {Sub_WFVM_Bot_WFVM : Sub_iFunctor (WFValueM_Bot (list (DType D))) WFVM}.

    Theorem eval_Pure_Sound_X :
      forall n Sigma gamma gamma' gamma'' e' e'',
        E_eqvC EQV_E gamma' gamma e'' e' ->
        forall (WF_gamma : forall n b, lookup gamma' n = Some b ->
          exists T, lookup gamma b = Some T)
        (WF_gamma'' : WF_EnvG (GammaTypContext := GammaTypContext) _ WFV gamma'' Sigma)
        (WF_gamma2 : List.length gamma = List.length gamma')
        (WF_gamma' : forall n b, lookup gamma' n = Some b -> b = n) 
        (WF_Sigma : Gamma Sigma = gamma) (T : DType D),
        typeof D (E (DType D)) MT (Typeof_E := Typeof_E) (proj1_sig e') = return_ T ->
        local (fun _ => gamma'') (bevalM n e'') = fail \/
        exists v : Value V, 
          local (fun _ => gamma'') (bevalM n e'') = return_ (M := ME) v /\
          WFValueC _ _ _ WFV Sigma v T.
    Proof.
      intros.      
      intros; eapply (ifold_ WFVM _ (ip_algebra (iPAlgebra := Pure_Sound_X_WFVM))
        _ (eval_soundness_X' (list (DType D)) _ WFVM n _ _ _ _ H Sigma WF_gamma WF_gamma2 WF_gamma' WF_Sigma)); eauto.
    Qed.

    Global Instance Pure_Sound_WFVM_base' :
      iPAlgebra Pure_Sound_X_Name Pure_Sound_X_P (WFValueM_base D V MT ME _ WFV).
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base with (WFV := WFV)
        (Fail_MT := _) (Monad_ME := _);
        try assumption; unfold Pure_Sound_X_P; intros; right.
      (* WFVM_Return' *)
      simpl; rewrite local_return; exists v; repeat split; simpl in *; auto.
      destruct H1 as [mt' mt'_eq]; subst.
      destruct mt'; try discriminate; injection H2; intros; subst.
      simpl in *; subst; auto.
      destruct s as [T0 T0_UP']; simpl in *.
      destruct T0; apply (WFV_proj1_b D V _ WFV _ _ H0); simpl; auto.
      (* WFVM_Untyped' *)
      simpl in H0; apply sym_eq in H0.
      destruct mt; discriminate.
    Qed.

    Context {FailEnvME : FailEnvironmentMonad (Env (Value V)) ME}.

    Global Instance Pure_Sound_WFVM_Bot :
      iPAlgebra Pure_Sound_X_Name Pure_Sound_X_P (WFValueM_Bot (list (DType D))).
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_Bot;
        try assumption; unfold Pure_Sound_X_P; intros.
    (* WFVM_fail *)
      left; simpl.
      apply local_fail.
    Qed.

  End Pure_Sound_X_Sec.

(*
    Context {eqv_eval_soundness_alg : forall typeof_rec eval_rec,
      iPAlgebra eqv_eval_SoundnessName
      (eqv_eval_alg_Soundness'_P
        typeof_rec eval_rec
        (f_algebra (FAlgebra := Typeof_E _))
        (f_algebra (FAlgebra := eval_E _))) (EQV_E _ _)}.

    Definition eqv_eval_soundness_P (i : eqv_i (typeofR D) nat) :=
      forall (gamma'' : Env _)
        (WF_gamma : forall n b, lookup (env_B _ _ i) n = Some b ->
          exists T, lookup (env_A _ _ i) b = Some T)
        (WF_gamma2 : List.length (env_A _ _ i) = List.length (env_B _ _ i))
        (WF_gamma' : forall n b, lookup (env_B _ _ i) n = Some b -> b = n)
        (WF_gamma'' : WF_Environment _ _ WFV gamma'' (env_A _ _ i)) T,
        typeof _ _ (proj1_sig (eqv_a _ _ i)) = Some T ->
        WFValueC _ _ WFV (eval (eval_E := eval_E)
          V _ (proj1_sig (eqv_b _ _ i)) gamma'') T.

    Variable (WF_MAlg_typeof : WF_MAlgebra Typeof_E).
    Variable (WF_MAlg_eval : WF_MAlgebra eval_E).

    Lemma eqv_eval_soundness' : forall gamma gamma' e' e'',
      E_eqvC EQV_E gamma gamma' e' e'' ->
      eqv_eval_soundness_P (mk_eqv_i _ _ gamma gamma' e' e'').
    Proof.
      intros; generalize (ifold_ (EQV_E _ _) _
        (ip_algebra (iPAlgebra := eqv_eval_soundness_alg
          (fun e => typeof D (E (typeofR D)) (proj1_sig e))
          (fun e => eval V (E nat) (proj1_sig e)))) (mk_eqv_i _ _ gamma gamma' e' e'') H).
      unfold eqv_eval_alg_Soundness'_P, eqv_eval_soundness_P; simpl;
        intros.
      revert H1.
      destruct e' as [e' e'_UP]; destruct e'' as [e'' e''_UP];
        simpl in *|-*.
      rewrite <- (@in_out_UP'_inverse _ _ e'' _).
      simpl; unfold typeof, eval, fold_, mfold, in_t.
      rewrite wf_malgebra; unfold mfold.
      unfold eval_alg_Soundness_P in H0.
      intros; eapply H0; unfold WF_eqv_environment_P.
      intro; rewrite (@in_out_UP'_inverse _ _ (proj1_sig e) (proj2_sig _)); reflexivity.
      intro; rewrite (@in_out_UP'_inverse _ _ (proj1_sig e) (proj2_sig _)); reflexivity.
      split; eauto.
      intros; simpl; unfold eval, mfold, in_t.
      rewrite wf_malgebra; eapply H2; eauto.
      rewrite <- (@in_out_UP'_inverse _ _ (proj1_sig (fst a)) (proj2_sig _)) in H3.
      simpl in H3; unfold typeof, mfold, in_t in H3.
      rewrite <- wf_malgebra;  apply H3.
      rewrite <- (@in_out_inverse _ _ e' _) in H1; unfold in_t in H1.
      simpl; rewrite <- wf_malgebra.
      simpl; unfold out_t_UP'.
      rewrite Fusion with (g := (fmap in_t)).
      apply H1.
      auto.
      intros; repeat rewrite fmap_fusion; reflexivity.
    Qed.

    Lemma eqv_eval_soundness : forall gamma gamma' e' e'',
        E_eqvC EQV_E gamma gamma' e' e'' ->
      forall (gamma'' : Env _)
        (WF_gamma : forall n b, lookup (gamma') n = Some b ->
          exists T, lookup (gamma) b = Some T)
        (WF_gamma2 : List.length (gamma) = List.length (gamma'))
        (WF_gamma' : forall n b, lookup (gamma') n = Some b -> b = n)
        (WF_gamma'' : WF_Environment _ _ WFV gamma'' (gamma)) T,
        typeof _ _ (proj1_sig e') = Some T ->
        WFValueC _ _ WFV (eval (eval_E := eval_E)
          V _ (proj1_sig (e'')) gamma'') T.
    Proof.
      intros; eapply eqv_eval_soundness'; eauto.
    Qed.

    Definition soundness_X'_P
      (typeof_rec : Exp (E (typeofR D)) -> typeofR D)
      (eval_rec : Exp (E nat) -> evalR V)
      (typeof_F : Mixin (Exp (E (typeofR D))) (E (typeofR D)) (typeofR D))
      (eval_F : Mixin (Exp (E nat)) (E nat) (evalR V))
      i :=
      forall (IH : forall (e : Exp _) (e' : Exp _)
        pb gamma'' (WF_gamma'' : WF_eqv_environment_P pb gamma'')
        T,
        E_eqvC EQV_E (fst pb) (snd pb) e e' ->
        typeof_rec e = Some T ->
        WFValueC _ _ WFV (eval_rec (in_t_UP' _ _ (out_t_UP' _ _ (proj1_sig e'))) gamma'') T),
      E_eqv EQV_E _ _ i /\
      eval_alg_Soundness_P D V (E nat) WFV
      _ WF_eqv_environment_P
      (E (typeofR D)) _ (env_A _ _ i, env_B _ _ i) typeof_rec eval_rec
      typeof_F eval_F
      (proj1_sig (eqv_a _ _ i), proj1_sig (eqv_b _ _ i))
      (conj (proj2_sig (eqv_a _ _ i)) (proj2_sig (eqv_b _ _ i))).

    Inductive soundness_XName : Set := soundness_Xname.

    Global Instance Lift_soundness_X_alg
      typeof_rec eval_rec typeof_alg eval_alg
      EQV_G {fun_EQV_G : iFunctor EQV_G}
      {EQV_G_EQV_Alg : iPAlgebra eqv_eval_SoundnessName
        (eqv_eval_alg_Soundness'_P typeof_rec eval_rec typeof_alg eval_alg) EQV_G} :
        iPAlgebra soundness_XName
        (soundness_X'_P
          typeof_rec eval_rec typeof_alg eval_alg) EQV_G.
      intros; econstructor; generalize (ip_algebra); unfold iAlgebra; intros.
      unfold soundness_X'_P; intros.
      assert (EQV_G (eqv_eval_alg_Soundness'_P typeof_rec eval_rec typeof_alg eval_alg) i).
      eapply ifmap; try eapply H0.
      intros; apply H1; apply IH.
      apply (H _ H1).
    Defined.

    Context {soundness_X_alg : forall eval_rec,
      iPAlgebra soundness_XName
      (soundness_X'_P
        (fun e => typeof _ _ (proj1_sig e)) eval_rec
        (f_algebra (FAlgebra := Typeof_E _))
        (f_algebra (FAlgebra := beval_E))) (EQV_E _ _)}.
    Variable Sub_WFV_Bot_WFV : Sub_iFunctor (WFValue_Bot _ _) WFV.

    Definition soundness_X_P (i : eqv_i (typeofR D) nat) :=
      forall n (gamma'' : Env _)
        (WF_gamma : forall n b, lookup (env_B _ _ i) n = Some b ->
          exists T, lookup (env_A _ _ i) b = Some T)
        (WF_gamma2 : List.length (env_A _ _ i) = List.length (env_B _ _ i))
        (WF_gamma' : forall n b, lookup (env_B _ _ i) n = Some b -> b = n)
        (WF_gamma'' : WF_Environment _ _ WFV gamma'' (env_A _ _ i)) T,
        typeof _ _ (proj1_sig (eqv_a _ _ i)) = Some T ->
        WFValueC _ _ WFV (beval V (E _) n (beval_E := beval_E)
          (eqv_b _ _ i) gamma'') T.

    Lemma soundness_X' :
      forall eval_rec gamma gamma' e' e'',
        E_eqvC EQV_E gamma gamma' e' e'' ->
        soundness_X'_P (fun e => typeof _ _ (proj1_sig e)) eval_rec
        (f_algebra (FAlgebra := Typeof_E _))
        (f_algebra (FAlgebra := beval_E)) (mk_eqv_i _ _ gamma gamma' e' e'').
    Proof.
      intros; apply (ifold_ (EQV_E _ _ )); try assumption.
      apply ip_algebra.
    Qed.

    Variable SV : (SubValue_i V -> Prop) -> SubValue_i V -> Prop.
    Variable funSV : iFunctor SV.
    Variable Sub_SV_Bot_SV : Sub_iFunctor (SubValue_Bot V) SV.
    Variable Sub_SV_refl_SV : Sub_iFunctor (SubValue_refl V) SV.

    Context {WF_Value_continous_alg :
      iPAlgebra WFV_ContinuousName (WF_Value_continuous_P D V WFV) SV}.
    Context {eval_continuous_Exp_E : PAlgebra EC_ExpName
      (sig (UP'_P (eval_continuous_Exp_P V (E _) SV))) (E nat)}.
    Context {WF_Ind_EC_Exp : WF_Ind eval_continuous_Exp_E}.

    Lemma soundness_X :
      forall n gamma gamma' gamma'' e' e'',
        E_eqvC EQV_E gamma gamma' e' e'' ->
        forall (WF_gamma : forall n b, lookup gamma' n = Some b ->
          exists T, lookup gamma b = Some T)
        (WF_gamma2 : List.length gamma = List.length gamma')
        (WF_gamma' : forall n b, lookup gamma' n = Some b -> b = n)
        (WF_gamma'' : WF_Environment _ _ WFV gamma'' gamma) T,
        typeof _ _ (proj1_sig e') = Some T ->
        WFValueC _ _ WFV (beval V (E _) n (beval_E := beval_E)
          e'' gamma'') T.
    Proof.
      induction n; simpl;
          intros; unfold beval; simpl in *|-*.
      apply (inject_i (subGF := Sub_WFV_Bot_WFV)); econstructor; eauto.
      generalize (soundness_X' (beval V (E _) n) _ _ _ _ H).
      unfold soundness_X'_P;
        unfold eval_alg_Soundness_P; simpl; intros.
      apply H1; auto.
      unfold beval; intros; erewrite bF_UP_in_out.
      instantiate (1 := proj2_sig _).
      destruct e'0; simpl; auto.
      destruct WF_gamma''0 as [WF_pb [WF_pb2 [WF_pb' WF_gamma''0]]].
      eapply IHn; eauto.
      intro; destruct e; unfold beval; erewrite bF_UP_in_out;
        auto.
      intro; rewrite <- (@in_out_UP'_inverse _ _ (proj1_sig e) (proj2_sig _)) at 1;
        reflexivity.
      repeat split; auto.
      intros.
      destruct a as [[a a_UP'] [a' a'_UP']].
      unfold beval; erewrite (@bF_UP_in_out _ _ _ _ _ _ _ a'_UP').
      apply (WF_Value_beval D V (E _) SV _ _ _ _ WFV n (S n) _ gamma''0); auto.
      apply Sub_Environment_refl; auto.
      unfold beval; simpl.
      simpl in H2.
      unfold beval in H2; apply H2.
      simpl in H3.
      unfold typeof in H3.
      rewrite <- (@in_out_UP'_inverse _ _ a a_UP') in H3.
      simpl in H3; unfold typeof, mfold, in_t in H3.
      rewrite <- wf_malgebra; apply H3.
      rewrite <- wf_malgebra.
      rewrite <- (@in_out_UP'_inverse _ _ _ (proj2_sig e')) in H0.
      simpl in H0; unfold typeof, mfold, in_t in H0.
      apply H0.
    Qed. *)


End PNames.

(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
